from .select_query_utilities import SELECTQueryParser, SELECTQueryExecutor
from .insert_query_utilities import INSERTQueryParser
from .update_query_utilities import UPDATEQueryParser, UPDATEQueryExecutor
from .delete_query_utilities import DELETEQueryParser, DELETEQueryExecutor



import pandas as pd

from typing import List

from mindsdb.integrations.libs.api_handler import APITable
from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions
from mindsdb.utilities import log

from mindsdb_sql_parser import ast
from mindsdb.integrations.utilities.handlers.query_utilities.select_query_utilities import SELECTQueryParser


logger = log.getLogger(__name__)

class InfluxDBTables(APITable):
    """InfluxDB Tables implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the InfluxDB "query" API endpoint
        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query
        Returns
        -------
        pd.DataFrame of particular InfluxDB table matching the query
        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        
        table_name=self.handler.connection_data['influxdb_table_name']
        select_statement_parser = SELECTQueryParser(
            query,
            "tables",
            self.get_columns()
        )
        selected_columns, where_conditions, order_by_conditions, _ = select_statement_parser.parse_query()
        
        try:
            selected_columns.remove("name")
            selected_columns.remove("tags")
        except Exception as e:
            logger.warn(e)

        formatted_query=self.get_select_query(table_name,selected_columns,where_conditions,order_by_conditions,query.limit)
        influxdb_tables_df  = self.handler.call_influxdb_tables(formatted_query)

        return influxdb_tables_df

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses
        Returns
        -------
        List[str]
            List of columns
        """
        
        dataframe = self.handler.call_influxdb_tables(f"SELECT * FROM {self.handler.connection_data['influxdb_table_name']} LIMIT 1")

        return list(dataframe.columns)
    def get_select_query(self,table_name,selected_columns, where_conditions, order_by_conditions, result_limit):
        """Gets Well formed Query
        Returns
        -------
        str
        """
        columns=", ".join([f'"{column}"' for column in selected_columns])
        query=f'SELECT {columns} FROM "{table_name}"'
        if(where_conditions is not None and len(where_conditions)>0):
            query+=" WHERE "
            query+=" AND ".join([f"{i[1]} {i[0]} {i[2]}" for i in where_conditions])
        if (order_by_conditions!={} and order_by_conditions['columns'] is not None and len(order_by_conditions['columns'])>0):
            query+=" ORDER BY "
            query+=", ".join([f'{column_name} {"ASC"if asc else "DESC"}' for column_name,asc in zip(order_by_conditions['columns'],order_by_conditions['ascending'])] )
        if(result_limit is not None):
            query+=f" LIMIT {result_limit}"
        query+=";"
        return query




__title__ = "MindsDB OpenBB handler"
__package_name__ = "mindsdb_openbb_handler"
__version__ = "0.0.1"
__description__ = "MindsDB handler for the OpenBB Platform"
__author__ = "Didier Rodrigues Lopes"
__github__ = "https://github.com/openbb-finance/openbbterminal"
__pypi__ = "https://pypi.org/project/openbb/"
__license__ = "MIT"
__copyright__ = "Copyright 2023 - OpenBB"



import os
try:
    import torch.multiprocessing as mp
except Exception:
    import multiprocessing as mp

from flask import Flask
from waitress import serve

from mindsdb.api.http.initialize import initialize_app
from mindsdb.interfaces.storage import db
from mindsdb.utilities import log
from mindsdb.utilities.config import config
from mindsdb.utilities.functions import init_lexer_parsers
from mindsdb.integrations.libs.ml_exec_base import process_cache

logger = log.getLogger(__name__)


def start(verbose, no_studio, app: Flask = None):
    server = os.environ.get('MINDSDB_DEFAULT_SERVER', 'waitress')
    db.init()
    init_lexer_parsers()

    if app is None:
        app = initialize_app(config, no_studio)

    port = config['api']['http']['port']
    host = config['api']['http']['host']

    process_cache.init()

    if server.lower() == "waitress":
        logger.debug("Serving HTTP app with waitress..")
        serve(
            app,
            host='*' if host in ('', '0.0.0.0') else host,
            port=port,
            threads=16,
            max_request_body_size=1073741824 * 10,
            inbuf_overflow=1073741824 * 10
        )
    elif server.lower() == "flask":
        logger.debug("Serving HTTP app with flask..")
        # that will 'disable access' log in console

        app.run(debug=False, port=port, host=host)
    elif server.lower() == 'gunicorn':
        try:
            from mindsdb.api.http.gunicorn_wrapper import StandaloneApplication
        except ImportError:
            logger.error(
                "Gunicorn server is not available by default. If you wish to use it, please install 'gunicorn'"
            )
            return

        def post_fork(arbiter, worker):
            db.engine.dispose()

        options = {
            'bind': f'{host}:{port}',
            'workers': mp.cpu_count(),
            'timeout': 600,
            'reuse_port': True,
            'preload_app': True,
            'post_fork': post_fork,
            'threads': 4
        }
        StandaloneApplication(app, options).run()



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    host={
        'type': ARG_TYPE.STR,
        'description': 'The host name or IP address of the Firebird server.'
    },
    database={
        'type': ARG_TYPE.STR,
        'description': """
            The database name to use when connecting with the Firebird server. NOTE: use double backslashes (\\) for the
            database path on a Windows machine.
        """
    },
    user={
        'type': ARG_TYPE.STR,
        'description': 'The user name used to authenticate with the Firebird server.'
    },
    password={
        'type': ARG_TYPE.PWD,
        'description': 'The password to authenticate the user with the Firebird server.',
        'secret': True
    }
)

connection_args_example = OrderedDict(
    host='localhost',
    database='/temp/test.db',
    user='sysdba',
    password='password'
)



import pandas as pd
from langchain.storage import InMemoryByteStore
from langchain_core.runnables import RunnableSerializable
from mindsdb.integrations.utilities.rag.pipelines.rag import LangChainRAGPipeline
from mindsdb.integrations.utilities.rag.settings import (
    RetrieverType,
    RAGPipelineModel
)
from mindsdb.integrations.utilities.rag.utils import documents_to_df
from mindsdb.utilities.log import getLogger
from langchain_text_splitters import RecursiveCharacterTextSplitter

logger = getLogger(__name__)

_retriever_strategies = {
    RetrieverType.VECTOR_STORE: lambda config: _create_pipeline_from_vector_store(config),
    RetrieverType.AUTO: lambda config: _create_pipeline_from_auto_retriever(config),
    RetrieverType.MULTI: lambda config: _create_pipeline_from_multi_retriever(config),
}


def _create_pipeline_from_vector_store(config: RAGPipelineModel) -> LangChainRAGPipeline:

    return LangChainRAGPipeline.from_retriever(
        config=config
    )


def _create_pipeline_from_auto_retriever(config: RAGPipelineModel) -> LangChainRAGPipeline:
    return LangChainRAGPipeline.from_auto_retriever(
        config=config
    )


def _create_pipeline_from_multi_retriever(config: RAGPipelineModel) -> LangChainRAGPipeline:

    if config.text_splitter is None:
        config.text_splitter = RecursiveCharacterTextSplitter(
            chunk_size=config.chunk_size, chunk_overlap=config.chunk_overlap
        )
    if config.parent_store is None:
        config.parent_store = InMemoryByteStore()

    return LangChainRAGPipeline.from_multi_vector_retriever(
        config=config
    )


def _process_documents_to_df(config: RAGPipelineModel) -> pd.DataFrame:
    return documents_to_df(config.content_column_name,
                           config.documents,
                           embedding_model=config.embedding_model,
                           with_embeddings=True)


def get_pipeline_from_retriever(config: RAGPipelineModel) -> RunnableSerializable:
    retriever_strategy = _retriever_strategies.get(config.retriever_type)
    if retriever_strategy:
        return retriever_strategy(config).with_returned_sources()
    else:
        raise ValueError(
            f'Invalid retriever type, must be one of: {list(_retriever_strategies.keys())}. Got {config.retriever_type}')


class RAG:
    def __init__(self, config: RAGPipelineModel):
        self.pipeline = get_pipeline_from_retriever(config)

    def __call__(self, question: str) -> dict:
        logger.info(f"Processing question using rag pipeline: {question}")
        result = self.pipeline.invoke(question)

        returned_sources = [docs.page_content for docs in result['context']]
        logger.info(f"retrieved context used to answer question: {returned_sources}")

        return result



"""add knowledge base

Revision ID: 309db3d07cf4
Revises: 6cb02dfd7f61
Create Date: 2023-09-20 13:48:39.422306

"""
import sqlalchemy as sa
from alembic import op

import mindsdb.interfaces.storage.db  # noqa

# revision identifiers, used by Alembic.
revision = "309db3d07cf4"
down_revision = "6cb02dfd7f61"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "knowledge_base",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("project_id", sa.Integer(), nullable=False),
        sa.Column("params", sa.JSON(), nullable=True),
        sa.Column("vector_database_id", sa.Integer(), nullable=True),
        sa.Column("vector_database_table", sa.String(), nullable=True),
        sa.Column("embedding_model_id", sa.Integer(), nullable=True),
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("updated_at", sa.DateTime(), nullable=True),
        sa.ForeignKeyConstraint(
            ["embedding_model_id"],
            ["predictor.id"],
            name="fk_knowledge_base_embedding_model_id",
        ),
        sa.ForeignKeyConstraint(
            ["vector_database_id"],
            ["integration.id"],
            name="fk_knowledge_base_vector_database_id",
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint(
            "name", "project_id", name="unique_knowledge_base_name_project_id"
        ),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("knowledge_base")
    # ### end Alembic commands ###



__title__ = 'MindsDB'
__package_name__ = 'mindsdb'
__version__ = '24.12.3.0'
__description__ = "MindsDB's AI SQL Server enables developers to build AI tools that need access to real-time data to perform their tasks"
__email__ = "jorge@mindsdb.com"
__author__ = 'MindsDB Inc'
__github__ = 'https://github.com/mindsdb/mindsdb'
__pypi__ = 'https://pypi.org/project/mindsdb/'
__license__ = 'SSPL v1'
__copyright__ = 'Copyright(c) 2018 MindsDB, Inc'



import os
import sys
import base64
import shutil
import ast
import time
import tempfile
import importlib
import threading
from pathlib import Path
from copy import deepcopy
from typing import Optional
from textwrap import dedent
from collections import OrderedDict
import inspect

from sqlalchemy import func

from mindsdb.interfaces.storage import db
from mindsdb.utilities.config import Config
from mindsdb.utilities.exception import EntityNotExistsError
from mindsdb.interfaces.storage.fs import FsStore, FileStorage, RESOURCE_GROUP
from mindsdb.interfaces.storage.model_fs import HandlerStorage
from mindsdb.interfaces.file.file_controller import FileController
from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.integrations.libs.base import BaseMLEngine
from mindsdb.integrations.libs.api_handler import APIHandler
from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE, HANDLER_TYPE
from mindsdb.interfaces.model.functions import get_model_records
from mindsdb.utilities.context import context as ctx
from mindsdb.utilities import log
from mindsdb.integrations.libs.ml_exec_base import BaseMLEngineExec
from mindsdb.integrations.libs.base import BaseHandler
import mindsdb.utilities.profiler as profiler

logger = log.getLogger(__name__)


class HandlersCache:
    """ Cache for data handlers that keep connections opened during ttl time from handler last use
    """

    def __init__(self, ttl: int = 60):
        """ init cache

            Args:
                ttl (int): time to live (in seconds) for record in cache
        """
        self.ttl = ttl
        self.handlers = {}
        self._lock = threading.RLock()
        self._stop_event = threading.Event()
        self.cleaner_thread = None

    def __del__(self):
        self._stop_clean()

    def _start_clean(self) -> None:
        """ start worker that close connections after ttl expired
        """
        if (
            isinstance(self.cleaner_thread, threading.Thread)
            and self.cleaner_thread.is_alive()
        ):
            return
        self._stop_event.clear()
        self.cleaner_thread = threading.Thread(target=self._clean)
        self.cleaner_thread.daemon = True
        self.cleaner_thread.start()

    def _stop_clean(self) -> None:
        """ stop clean worker
        """
        self._stop_event.set()

    def set(self, handler: DatabaseHandler):
        """ add (or replace) handler in cache

            Args:
                handler (DatabaseHandler)
        """
        with self._lock:
            try:
                # If the handler is defined to be thread safe, set 0 as the last element of the key, otherwise set the thrad ID.
                key = (handler.name, ctx.company_id, 0 if getattr(handler, 'thread_safe', False) else threading.get_native_id())
                handler.connect()
                self.handlers[key] = {
                    'handler': handler,
                    'expired_at': time.time() + self.ttl
                }
            except Exception:
                pass
            self._start_clean()

    def get(self, name: str) -> Optional[DatabaseHandler]:
        """ get handler from cache by name

            Args:
                name (str): handler name

            Returns:
                DatabaseHandler
        """
        with self._lock:
            # If the handler is not thread safe, the thread ID will be assigned to the last element of the key.
            key = (name, ctx.company_id, threading.get_native_id())
            if key not in self.handlers:
                # If the handler is thread safe, a 0 will be assigned to the last element of the key.
                key = (name, ctx.company_id, 0)
            if (
                key not in self.handlers
                or self.handlers[key]['expired_at'] < time.time()
            ):
                return None
            self.handlers[key]['expired_at'] = time.time() + self.ttl
            return self.handlers[key]['handler']

    def delete(self, name: str) -> None:
        """ delete handler from cache

            Args:
                name (str): handler name
        """
        with self._lock:
            key = (name, ctx.company_id, threading.get_native_id())
            if key in self.handlers:
                try:
                    self.handlers[key].disconnect()
                except Exception:
                    pass
                del self.handlers[key]
            if len(self.handlers) == 0:
                self._stop_clean()

    def _clean(self) -> None:
        """ worker that delete from cache handlers that was not in use for ttl
        """
        while self._stop_event.wait(timeout=3) is False:
            with self._lock:
                for key in list(self.handlers.keys()):
                    if (
                        self.handlers[key]['expired_at'] < time.time()
                        and sys.getrefcount(self.handlers[key]) == 2    # returned ref count is always 1 higher
                    ):
                        try:
                            self.handlers[key].disconnect()
                        except Exception:
                            pass
                        del self.handlers[key]
                if len(self.handlers) == 0:
                    self._stop_event.set()


class IntegrationController:
    @staticmethod
    def _is_not_empty_str(s):
        return isinstance(s, str) and len(s) > 0

    def __init__(self):
        self._import_lock = threading.Lock()
        self._load_handler_modules()
        self.handlers_cache = HandlersCache()

    def _add_integration_record(self, name, engine, connection_args):
        integration_record = db.Integration(
            name=name,
            engine=engine,
            data=connection_args or {},
            company_id=ctx.company_id
        )
        db.session.add(integration_record)
        db.session.commit()
        return integration_record.id

    def add(self, name, engine, connection_args):

        logger.debug(
            "%s: add method calling name=%s, engine=%s, connection_args=%s, company_id=%s",
            self.__class__.__name__, name, engine, connection_args, ctx.company_id
        )
        handler_meta = self.get_handler_meta(engine)

        accept_connection_args = handler_meta.get('connection_args')
        logger.debug("%s: accept_connection_args - %s", self.__class__.__name__, accept_connection_args)

        files_dir = None
        if accept_connection_args is not None and connection_args is not None:
            for arg_name, arg_value in connection_args.items():
                if (
                    arg_name in accept_connection_args
                    and accept_connection_args[arg_name]['type'] == ARG_TYPE.PATH
                ):
                    if files_dir is None:
                        files_dir = tempfile.mkdtemp(prefix='mindsdb_files_')
                    shutil.copy(arg_value, files_dir)
                    connection_args[arg_name] = Path(arg_value).name

        integration_id = self._add_integration_record(name, engine, connection_args)

        if files_dir is not None:
            store = FileStorage(
                resource_group=RESOURCE_GROUP.INTEGRATION,
                resource_id=integration_id,
                sync=False
            )
            store.add(files_dir, '')
            store.push()

        if handler_meta.get('type') == HANDLER_TYPE.ML:
            ml_handler = self.get_ml_handler(name)
            ml_handler.create_engine(connection_args, integration_id)

        return integration_id

    def modify(self, name, data):
        self.handlers_cache.delete(name)
        integration_record = self._get_integration_record(name)
        old_data = deepcopy(integration_record.data)
        for k in old_data:
            if k not in data:
                data[k] = old_data[k]

        integration_record.data = data
        db.session.commit()

    def delete(self, name):
        if name in ('files', 'lightwood'):
            raise Exception('Unable to drop: is system database')

        self.handlers_cache.delete(name)

        # check permanent integration
        if name in self.handler_modules:
            handler = self.handler_modules[name]

            if getattr(handler, 'permanent', False) is True:
                raise Exception('Unable to drop: is permanent integration')

        integration_record = self._get_integration_record(name)

        # if this is ml engine
        engine_models = get_model_records(ml_handler_name=name, deleted_at=None)
        active_models = [m.name for m in engine_models if m.deleted_at is None]
        if len(active_models) > 0:
            raise Exception(f'Unable to drop ml engine with active models: {active_models}')

        # check linked KBs
        kb = db.KnowledgeBase.query.filter_by(vector_database_id=integration_record.id).first()
        if kb is not None:
            raise Exception(f'Unable to drop, integration is used by knowledge base: {kb.name}')

        # check linked predictors
        models = get_model_records()
        for model in models:
            if (
                model.data_integration_ref is not None
                and model.data_integration_ref.get('type') == 'integration'
                and isinstance(model.data_integration_ref.get('id'), int)
                and model.data_integration_ref['id'] == integration_record.id
            ):
                model.data_integration_ref = None

        # unlink deleted models
        for model in engine_models:
            if model.deleted_at is not None:
                model.integration_id = None

        db.session.delete(integration_record)
        db.session.commit()

    def _get_integration_record_data(self, integration_record, show_secrets=True):
        if (
            integration_record is None
            or integration_record.data is None
            or isinstance(integration_record.data, dict) is False
        ):
            return None
        data = deepcopy(integration_record.data)

        bundle_path = data.get('secure_connect_bundle')
        mysql_ssl_ca = data.get('ssl_ca')
        mysql_ssl_cert = data.get('ssl_cert')
        mysql_ssl_key = data.get('ssl_key')
        if (
            data.get('type') in ('mysql', 'mariadb')
            and (
                self._is_not_empty_str(mysql_ssl_ca)
                or self._is_not_empty_str(mysql_ssl_cert)
                or self._is_not_empty_str(mysql_ssl_key)
            )
            or data.get('type') in ('cassandra', 'scylla')
            and bundle_path is not None
        ):
            fs_store = FsStore()
            integrations_dir = Config()['paths']['integrations']
            folder_name = f'integration_files_{integration_record.company_id}_{integration_record.id}'
            fs_store.get(
                folder_name,
                base_dir=integrations_dir
            )

        handler_meta = self.get_handler_metadata(integration_record.engine)
        integration_type = None
        if isinstance(handler_meta, dict):
            # in other cases, the handler directory is likely not exist.
            integration_type = handler_meta.get('type')

        if show_secrets is False and handler_meta is not None:
            connection_args = handler_meta.get('connection_args', None)
            if isinstance(connection_args, dict):
                if integration_type == HANDLER_TYPE.DATA:
                    for key, value in connection_args.items():
                        if key in data and value.get('secret', False) is True:
                            data[key] = '******'
                elif integration_type == HANDLER_TYPE.ML:
                    creation_args = connection_args.get('creation_args')
                    if isinstance(creation_args, dict):
                        for key, value in creation_args.items():
                            if key in data and value.get('secret', False) is True:
                                data[key] = '******'
                else:
                    raise ValueError(f'Unexpected handler type: {integration_type}')
            else:
                # region obsolete, del in future
                if 'password' in data:
                    data['password'] = None
                if (
                    data.get('type') == 'redis'
                    and isinstance(data.get('connection'), dict)
                    and 'password' in data['connection']
                ):
                    data['connection'] = None
                # endregion

        class_type, permanent = None, False
        if handler_meta is not None:
            class_type = handler_meta.get('class_type')
            permanent = handler_meta.get('permanent', False)

        return {
            'id': integration_record.id,
            'name': integration_record.name,
            'type': integration_type,
            'class_type': class_type,
            'engine': integration_record.engine,
            'permanent': permanent,
            'date_last_update': deepcopy(integration_record.updated_at),  # to del ?
            'connection_data': data
        }

    def get_by_id(self, integration_id, show_secrets=True):
        integration_record = (
            db.session.query(db.Integration)
            .filter_by(company_id=ctx.company_id, id=integration_id)
            .first()
        )
        return self._get_integration_record_data(integration_record, show_secrets)

    def get(self, name, show_secrets=True, case_sensitive=False):
        try:
            integration_record = self._get_integration_record(name, case_sensitive)
        except EntityNotExistsError:
            return None
        return self._get_integration_record_data(integration_record, show_secrets)

    @staticmethod
    def _get_integration_record(name: str, case_sensitive: bool = False) -> db.Integration:
        """Get integration record by name

        Args:
            name (str): name of the integration
            case_sensitive (bool): should search be case sensitive or not

        Retruns:
            db.Integration
        """
        if case_sensitive:
            integration_records = db.session.query(db.Integration).filter_by(
                company_id=ctx.company_id,
                name=name
            ).all()
            if len(integration_records) > 1:
                raise Exception(f"There is {len(integration_records)} integrations with name '{name}'")
            if len(integration_records) == 0:
                raise EntityNotExistsError(f"There is no integration with name '{name}'")
            integration_record = integration_records[0]
        else:
            integration_record = db.session.query(db.Integration).filter(
                (db.Integration.company_id == ctx.company_id)
                & (func.lower(db.Integration.name) == func.lower(name))
            ).first()
            if integration_record is None:
                raise EntityNotExistsError(f"There is no integration with name '{name}'")

        return integration_record

    def get_all(self, show_secrets=True):
        integration_records = db.session.query(db.Integration).filter_by(company_id=ctx.company_id).all()
        integration_dict = {}
        for record in integration_records:
            if record is None or record.data is None:
                continue
            integration_dict[record.name] = self._get_integration_record_data(record, show_secrets)
        return integration_dict

    def _make_handler_args(self, name: str, handler_type: str, connection_data: dict, integration_id: int = None,
                           file_storage: FileStorage = None, handler_storage: HandlerStorage = None):
        handler_args = dict(
            name=name,
            integration_id=integration_id,
            connection_data=connection_data,
            file_storage=file_storage,
            handler_storage=handler_storage
        )

        if handler_type == 'files':
            handler_args['file_controller'] = FileController()
        elif self.handler_modules.get(handler_type, False).type == HANDLER_TYPE.ML:
            handler_args['handler_controller'] = self
            handler_args['company_id'] = ctx.company_id

        return handler_args

    def create_tmp_handler(self, name: str, engine: str, connection_args: dict) -> dict:
        """Create temporary handler, mostly for testing connections

        Args:
            name (str): Integration  name
            engine (str): Integration engine name
            connection_args (dict): Connection arguments

        Returns:
            HandlerClass: Handler class instance
        """
        integration_id = int(time.time() * 10000)

        file_storage = FileStorage(
            resource_group=RESOURCE_GROUP.INTEGRATION,
            resource_id=integration_id,
            root_dir='tmp',
            sync=False
        )
        handler_storage = HandlerStorage(integration_id, root_dir='tmp', is_temporal=True)

        handler_meta = self.get_handler_meta(engine)
        if handler_meta is None:
            raise ImportError(f"Handler '{engine}' does not exist")
        if handler_meta["import"]["success"] is False:
            raise ImportError(f"Handler '{engine}' cannot be imported: {handler_meta['import']['error_message']}")
        HandlerClass = self.handler_modules[engine].Handler

        handler_args = self._make_handler_args(
            name=name,
            handler_type=engine,
            connection_data=connection_args,
            integration_id=integration_id,
            file_storage=file_storage,
            handler_storage=handler_storage,
        )
        handler = HandlerClass(**handler_args)
        return handler

    def copy_integration_storage(self, integration_id_from, integration_id_to):
        storage_from = HandlerStorage(integration_id_from)
        root_path = ''

        if storage_from.is_empty():
            return None
        folder_from = storage_from.folder_get(root_path)

        storage_to = HandlerStorage(integration_id_to)
        folder_to = storage_to.folder_get(root_path)

        shutil.copytree(folder_from, folder_to, dirs_exist_ok=True)
        storage_to.folder_sync(root_path)

    def get_ml_handler(self, name: str, case_sensitive: bool = False) -> BaseMLEngine:
        """Get ML handler by name
        Args:
            name (str): name of the handler
            case_sensitive (bool): should case be taken into account when searching by name

        Returns:
            BaseMLEngine
        """
        integration_record = self._get_integration_record(name, case_sensitive)
        integration_engine = integration_record.engine

        integration_meta = self.get_handler_meta(integration_engine)
        if integration_meta is None:
            raise Exception(f"Handler '{name}' does not exists")

        if integration_meta.get('type') != HANDLER_TYPE.ML:
            raise Exception(f"Handler '{name}' must be ML type")

        logger.info(
            f"{self.__class__.__name__}.get_handler: create a ML client "
            + f"{integration_record.name}/{integration_record.id}"
        )
        handler = BaseMLEngineExec(
            name=integration_record.name,
            integration_id=integration_record.id,
            handler_module=self.handler_modules[integration_engine]
        )

        return handler

    @profiler.profile()
    def get_data_handler(self, name: str, case_sensitive: bool = False, connect=True) -> BaseHandler:
        """Get DATA handler (DB or API) by name

        Args:
            name (str): name of the handler
            case_sensitive (bool): should case be taken into account when searching by name

        Returns:
            BaseHandler: data handler
        """
        handler = self.handlers_cache.get(name)
        if handler is not None:
            return handler

        integration_record = self._get_integration_record(name, case_sensitive)
        integration_engine = integration_record.engine

        integration_meta = self.get_handler_meta(integration_engine)

        if integration_meta is None:
            raise Exception(f"Handler '{name}' does not exist")

        if integration_meta.get('type') != HANDLER_TYPE.DATA:
            raise Exception(f"Handler '{name}' must be DATA type")

        integration_data = self._get_integration_record_data(integration_record, True)
        if integration_data is None:
            raise Exception(f"Can't find integration_record for handler '{name}'")
        connection_data = integration_data.get('connection_data', {})
        logger.debug(
            "%s.get_handler: connection_data=%s, engine=%s",
            self.__class__.__name__,
            connection_data, integration_engine
        )

        if integration_meta["import"]["success"] is False:
            msg = dedent(f'''\
                Handler '{integration_engine}' cannot be used. Reason is:
                    {integration_meta['import']['error_message']}
            ''')
            is_cloud = Config().get('cloud', False)
            if is_cloud is False:
                msg += dedent(f'''

                If error is related to missing dependencies, then try to run command in shell and restart mindsdb:
                    pip install mindsdb[{integration_engine}]
                ''')
            logger.debug(msg)
            raise Exception(msg)

        connection_args = integration_meta.get('connection_args')
        logger.debug("%s.get_handler: connection args - %s", self.__class__.__name__, connection_args)

        file_storage = FileStorage(
            resource_group=RESOURCE_GROUP.INTEGRATION,
            resource_id=integration_record.id,
            sync=True,
        )
        handler_storage = HandlerStorage(integration_record.id)

        if isinstance(connection_args, (dict, OrderedDict)):
            files_to_get = {
                arg_name: arg_value for arg_name, arg_value in connection_data.items()
                if arg_name in connection_args and connection_args.get(arg_name)['type'] == ARG_TYPE.PATH
            }
            if len(files_to_get) > 0:

                for file_name, file_path in files_to_get.items():
                    connection_data[file_name] = file_storage.get_path(file_path)

        handler_ars = self._make_handler_args(
            name=name,
            handler_type=integration_engine,
            connection_data=connection_data,
            integration_id=integration_data['id'],
            file_storage=file_storage,
            handler_storage=handler_storage
        )

        HandlerClass = self.handler_modules[integration_engine].Handler
        handler = HandlerClass(**handler_ars)
        if connect:
            self.handlers_cache.set(handler)

        return handler

    def reload_handler_module(self, handler_name):
        importlib.reload(self.handler_modules[handler_name])
        try:
            handler_meta = self._get_handler_meta(handler_name)
        except Exception as e:
            handler_meta = self.handlers_import_status[handler_name]
            handler_meta['import']['success'] = False
            handler_meta['import']['error_message'] = str(e)

        self.handlers_import_status[handler_name] = handler_meta

    def _read_dependencies(self, path):
        dependencies = []
        requirements_txt = Path(path).joinpath('requirements.txt')
        if requirements_txt.is_file():
            with open(str(requirements_txt), 'rt') as f:
                dependencies = [x.strip(' \t\n') for x in f.readlines()]
                dependencies = [x for x in dependencies if len(x) > 0]
        return dependencies

    def _get_handler_meta(self, handler_name):

        module = self.handler_modules[handler_name]

        handler_dir = Path(module.__path__[0])
        handler_folder_name = handler_dir.name

        import_error = getattr(module, 'import_error', None)
        handler_meta = self.handlers_import_status[handler_name]
        handler_meta['import']['success'] = import_error is None
        handler_meta['version'] = module.version
        handler_meta['thread_safe'] = getattr(module, 'thread_safe', False)

        if import_error is not None:
            handler_meta['import']['error_message'] = str(import_error)

        handler_type = getattr(module, 'type', None)
        handler_class = None
        if hasattr(module, 'Handler') and inspect.isclass(module.Handler):
            handler_class = module.Handler
            if issubclass(handler_class, BaseMLEngine):
                handler_meta['class_type'] = 'ml'
            elif issubclass(handler_class, DatabaseHandler):
                handler_meta['class_type'] = 'sql'
            if issubclass(handler_class, APIHandler):
                handler_meta['class_type'] = 'api'

        if handler_type == HANDLER_TYPE.ML:
            # for ml engines, patch the connection_args from the argument probing
            if handler_class:
                try:
                    prediction_args = handler_class.prediction_args()
                    creation_args = getattr(module, 'creation_args', handler_class.creation_args())
                    connection_args = {
                        "prediction": prediction_args,
                        "creation_args": creation_args
                    }
                    setattr(module, 'connection_args', connection_args)
                    logger.debug("Patched connection_args for %s", handler_folder_name)
                except Exception as e:
                    # do nothing
                    logger.debug("Failed to patch connection_args for %s, reason: %s", handler_folder_name, str(e))

        module_attrs = [attr for attr in [
            'connection_args_example',
            'connection_args',
            'description',
            'type',
            'title'
        ] if hasattr(module, attr)]

        for attr in module_attrs:
            handler_meta[attr] = getattr(module, attr)

        # endregion
        if hasattr(module, 'permanent'):
            handler_meta['permanent'] = module.permanent
        else:
            if handler_meta.get('name') in ('files', 'views', 'lightwood'):
                handler_meta['permanent'] = True
            else:
                handler_meta['permanent'] = False

        return handler_meta

    def _get_handler_icon(self, handler_dir, icon_path):
        icon = {}
        try:
            icon_path = handler_dir.joinpath(icon_path)
            icon_type = icon_path.name[icon_path.name.rfind('.') + 1:].lower()

            if icon_type == 'svg':
                with open(str(icon_path), 'rt') as f:
                    icon['data'] = f.read()
            else:
                with open(str(icon_path), 'rb') as f:
                    icon['data'] = base64.b64encode(f.read()).decode('utf-8')

            icon['name'] = icon_path.name
            icon['type'] = icon_type

        except Exception as e:
            logger.error(f'Error reading icon for {handler_dir}, {e}!')
        return icon

    def _load_handler_modules(self):
        mindsdb_path = Path(importlib.util.find_spec('mindsdb').origin).parent
        handlers_path = mindsdb_path.joinpath('integrations/handlers')

        # edge case: running from tests directory, find_spec finds the base folder instead of actual package
        if not os.path.isdir(handlers_path):
            mindsdb_path = Path(importlib.util.find_spec('mindsdb').origin).parent.joinpath('mindsdb')
            handlers_path = mindsdb_path.joinpath('integrations/handlers')

        self.handler_modules = {}
        self.handlers_import_status = {}
        for handler_dir in handlers_path.iterdir():
            if handler_dir.is_dir() is False or handler_dir.name.startswith('__'):
                continue

            handler_info = self._get_handler_info(handler_dir)
            if 'name' not in handler_info:
                continue
            handler_name = handler_info['name']
            dependencies = self._read_dependencies(handler_dir)
            handler_meta = {
                'path': handler_dir,
                'import': {
                    'success': None,
                    'error_message': None,
                    'folder': handler_dir.name,
                    'dependencies': dependencies,
                },
                'name': handler_name,
                'permanent': handler_info.get('permanent', False),
                'connection_args': handler_info.get('connection_args', None),
                'class_type': handler_info.get('class_type', None),
                'type': handler_info.get('type')
            }
            if 'icon_path' in handler_info:
                icon = self._get_handler_icon(handler_dir, handler_info['icon_path'])
                if icon:
                    handler_meta['icon'] = icon
            self.handlers_import_status[handler_name] = handler_meta

    def _get_connection_args(self, args_file: Path, param_name: str) -> dict:
        """
        Extract connection args dict from connection args file of a handler

        :param args_file: path to file with connection args
        :param param_name: the name of variable which contains connection args
        :return: extracted connection arguments
        """

        code = ast.parse(args_file.read_text())

        args = {}
        for item in code.body:
            if not isinstance(item, ast.Assign):
                continue
            if not item.targets[0].id == param_name:
                continue
            if hasattr(item.value, 'keywords'):
                for keyword in item.value.keywords:
                    name = keyword.arg
                    params = keyword.value
                    if isinstance(params, ast.Dict):
                        # get dict value
                        info = {}
                        for i, k in enumerate(params.keys):
                            if not isinstance(k, ast.Constant):
                                continue
                            v = params.values[i]
                            if isinstance(v, ast.Constant):
                                v = v.value
                            elif isinstance(v, ast.Attribute):
                                # assume it is ARG_TYPE
                                v = getattr(ARG_TYPE, v.attr, None)
                            else:
                                v = None
                            info[k.value] = v
                    args[name] = info
        return args

    def _get_base_class_type(self, code, handler_dir: Path) -> Optional[str]:
        """
        Find base class of data handler: sql or api
        It tries to find import inside try-except of init file
        They parsed this import in order to find base class of data handler

        :param code: parsed code of __init__ file of a handler
        :param handler_dir: folder of a handler
        :return: base class type
        """

        module_file = None
        for block in code.body:
            if not isinstance(block, ast.Try):
                continue
            for item in block.body:
                if isinstance(item, ast.ImportFrom):
                    module_file = item.module
                    break
        if module_file is None:
            return

        path = handler_dir / f'{module_file}.py'

        if not path.exists():
            return
        code = ast.parse(path.read_text())
        # find base class of handler.
        #  TODO trace inheritance (is used only for sql handler)
        for item in code.body:
            if isinstance(item, ast.ClassDef):
                bases = [base.id for base in item.bases]
                if 'APIHandler' in bases:
                    return 'api'
        return 'sql'

    def _get_handler_info(self, handler_dir: Path) -> dict:
        """
        Get handler info without importing it
        :param handler_dir: folder of handler
        :return: Extracted params:
          - defined constants in init file
          - connection arguments
        """

        init_file = handler_dir / '__init__.py'
        if not init_file.exists():
            return {}
        code = ast.parse(init_file.read_text())

        info = {}
        for item in code.body:
            if not isinstance(item, ast.Assign):
                continue
            if isinstance(item.targets[0], ast.Name):
                name = item.targets[0].id
                if isinstance(item.value, ast.Constant):
                    info[name] = item.value.value
                if isinstance(item.value, ast.Attribute) and name == 'type':
                    if item.value.attr == 'ML':
                        info[name] = HANDLER_TYPE.ML
                        info['class_type'] = 'ml'
                    else:
                        info[name] = HANDLER_TYPE.DATA
                        info['class_type'] = self._get_base_class_type(code, handler_dir) or 'sql'

        # connection args
        if info['type'] == HANDLER_TYPE.ML:
            args_file = handler_dir / 'creation_args.py'
            if args_file.exists():
                info['connection_args'] = {
                    "prediction": {},
                    "creation_args": self._get_connection_args(args_file, 'creation_args')
                }
        else:
            args_file = handler_dir / 'connection_args.py'
            if args_file.exists():
                info['connection_args'] = self._get_connection_args(args_file, 'connection_args')

        return info

    def import_handler(self, handler_name: str, base_import: str = None):
        with self._import_lock:
            handler_meta = self.handlers_import_status[handler_name]
            handler_dir = handler_meta['path']

            handler_folder_name = str(handler_dir.name)
            if base_import is None:
                base_import = 'mindsdb.integrations.handlers.'

            try:
                handler_module = importlib.import_module(f'{base_import}{handler_folder_name}')
                self.handler_modules[handler_name] = handler_module
                handler_meta = self._get_handler_meta(handler_name)
            except Exception as e:
                handler_meta['import']['success'] = False
                handler_meta['import']['error_message'] = str(e)

            self.handlers_import_status[handler_meta['name']] = handler_meta
            return handler_meta

    def get_handlers_import_status(self):
        # tries to import all not imported yet

        result = {}
        for handler_name in list(self.handlers_import_status.keys()):
            handler_meta = self.get_handler_meta(handler_name)
            result[handler_name] = handler_meta

        return result

    def get_handlers_metadata(self):
        return self.handlers_import_status

    def get_handler_metadata(self, handler_name):
        # returns metadata
        return self.handlers_import_status.get(handler_name)

    def get_handler_meta(self, handler_name):
        # returns metadata and tries to import it
        handler_meta = self.handlers_import_status.get(handler_name)
        if handler_meta is None:
            return
        if handler_meta["import"]["success"] is None:
            handler_meta = self.import_handler(handler_name)
        return handler_meta

    def get_handler_module(self, handler_name):
        handler_meta = self.get_handler_meta(handler_name)
        if handler_meta is None:
            return
        if handler_meta["import"]["success"]:
            return self.handler_modules[handler_name]

    def create_permanent_integrations(self):
        for (
            integration_name,
            handler,
        ) in self.get_handlers_metadata().items():
            if handler.get("permanent"):
                integration_meta = integration_controller.get(name=integration_name)
                if integration_meta is None:
                    integration_record = db.Integration(
                        name=integration_name,
                        data={},
                        engine=integration_name,
                        company_id=None,
                    )
                    db.session.add(integration_record)
        db.session.commit()


integration_controller = IntegrationController()



"""add_agents_table

Revision ID: d44ab65a6a35
Revises: ad04ee0bd385
Create Date: 2023-09-06 11:32:08.777661

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'd44ab65a6a35'
down_revision = 'ad04ee0bd385'
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        'agents',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('company_id', sa.Integer(), nullable=True),
        sa.Column('user_class', sa.Integer(), nullable=True),
        sa.Column('name', sa.String(), nullable=False),
        sa.Column('project_id', sa.Integer(), nullable=False),
        sa.Column('model_name', sa.String(), nullable=False),
        sa.Column('params', sa.JSON(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.PrimaryKeyConstraint('id'))


def downgrade():
    op.drop_table('agents')



import pandas as pd
from mindsdb.utilities import log
from mindsdb_sql_parser import parse_sql
from mindsdb.integrations.libs.api_handler import APIHandler, FuncParser
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)

from datetime import datetime
import plaid
from plaid.api import plaid_api
from plaid.model.accounts_balance_get_request import AccountsBalanceGetRequest
from plaid.model.accounts_balance_get_request_options import AccountsBalanceGetRequestOptions
from plaid.model.transactions_get_request import TransactionsGetRequest
from plaid.model.transactions_get_request_options import TransactionsGetRequestOptions
from .plaid_tables import BalanceTable, TransactionTable
from .utils import parse_transaction


PLAID_ENV = {
    'production': plaid.Environment.Production,
    'development': plaid.Environment.Development,
    'sandbox': plaid.Environment.Sandbox,
}

logger = log.getLogger(__name__)

class PlaidHandler(APIHandler):
    '''A class for handling connections and interactions with the Plaid API.

    Attributes:
        plaid_env (str): Enviroment used by user [ 'sandbox'(default) OR 'development' OR 'production' ].
        client_id (str): Your Plaid API client_id. 
        secret (str): Your Plaid API secret
        access_token (str): The access token for the Plaid account.
    '''

    def __init__(self, name=None, **kwargs):
        super().__init__(name)

        args = kwargs.get('connection_data', {})

        self.plaid_config = plaid.Configuration(
            host=PLAID_ENV[args.get('plaid_env', 'sandbox')],
            api_key={
                'clientId': args.get('client_id'),
                'secret': args.get('secret')
            }
        )

        self.access_token = args.get('access_token')

        self.api = None
        self.is_connected = False

        balance = BalanceTable(self)
        transactions = TransactionTable(self)
        self._register_table('balance', balance)
        self._register_table('transactions', transactions)

    def connect(self):
        '''Authenticate with the Plaid API using the API keys and secrets stored in the `plaid_env`, `client_id`, `secret` , and `access_token`  attributes.'''  # noqa

        if self.is_connected is True:
            return self.api

        api_client = plaid.ApiClient(self.plaid_config)
        self.api = plaid_api.PlaidApi(api_client)
        self.is_connected = True
        return self.api

    def check_connection(self) -> StatusResponse:
        '''It evaluates if the connection with Plaid API is alive and healthy.
        Returns:
            HandlerStatusResponse
        '''

        response = StatusResponse(False)

        try:
            api = self.connect()
            api.accounts_balance_get(AccountsBalanceGetRequest(
                access_token=self.access_token)
            )
            response.success = True

        except Exception as e:
            response.error_message = f'Error connecting to Plaid api: {e}. '
            logger.error(response.error_message)

        if response.success is False and self.is_connected is True:
            self.is_connected = False

        return response

    def native_query(self, query_string: str = None):
        '''It parses any native statement string and acts upon it (for example, raw syntax commands).
        Args:
        query (Any): query in native format (str for sql databases,
            dict for mongo, api's json etc)
        Returns:
            HandlerResponse
        '''

        method_name, params = FuncParser().from_string(query_string)
        df = self.call_plaid_api(method_name, params)
        return Response(
            RESPONSE_TYPE.TABLE,
            data_frame=df
        )

    def call_plaid_api(self, method_name: str = None, params: dict = {}):
        '''Receive query as AST (abstract syntax tree) and act upon it somehow.
        Args:
            query (ASTNode): sql query represented as AST. May be any kind of query: SELECT, INSERT, DELETE, etc
        Returns:
            DataFrame
        '''

        result=pd.DataFrame()
        if method_name == 'get_balance':
            result = self.get_balance(params=params)
            result = BalanceTable(self).filter_columns(result=result)

        elif method_name == 'get_transactions':
            result = self.get_transactions(params=params)
            result = TransactionTable(self).filter_columns(result=result)

        return result

    def get_balance(self, params=None):
        '''Filters data from Plaid API's balance endpoint and returns a DataFrame with the required information.

        Args:
            params (dict, optional): A dictionary of options to be passed to the Plaid API.

        Returns:
            pandas.DataFrame: A DataFrame containing the filtered data.
        '''

        self.connect()
        if params.get('last_updated_datetime') is not None:
            options = AccountsBalanceGetRequestOptions(
                min_last_updated_datetime=datetime.strptime( 
                    params.get('last_updated_datetime')
                    )
            )

            response = self.api.accounts_balance_get(
                AccountsBalanceGetRequest(
                    access_token=self.access_token,
                    options=options
                )
            )
        else:
            response = self.api.accounts_balance_get(
                AccountsBalanceGetRequest(access_token=self.access_token)
            )

        messages = []
        for obj in response['accounts']:
            message_dict = {}
            for i in obj.to_dict().keys():
                if i.startswith('account_'):
                    message_dict[i] = obj[i]
                elif i == 'balances':
                    dict_obj = obj[i].to_dict()
                    for j in dict_obj.keys():
                        message_dict[f'balance_{j}'] = dict_obj[j]
                else:
                    message_dict[f'account_{i}'] = obj[i]
            messages.append(message_dict)
        df = pd.DataFrame(messages)

        return df

    def get_transactions(self, params={}):
        '''
        Filters data from Plaid API's transaction endpoint and returns a DataFrame with the required information.
        Args:
            params (dict, optional): A dictionary of options to be passed to the Plaid API.

        Returns:
            pandas.DataFrame: A DataFrame containing the filtered data.
        '''

        self.connect()
        if params.get('start_date') and params.get('end_date'):
            start_date = datetime.strptime(params.get('start_date'), '%Y-%m-%d').date()
            end_date = datetime.strptime(params.get('end_date'), '%Y-%m-%d').date()
        else:
            raise Exception('start_date and end_date is required in format YYYY-MM-DD ')
      
        request = TransactionsGetRequest(
            access_token=self.access_token,
            start_date=start_date,
            end_date=end_date,
            options=TransactionsGetRequestOptions()
        )

        response = self.api.transactions_get(request)
        transactions = parse_transaction(response['transactions'])

        # Manipulate the count and offset parameters to paginate
        # transactions and retrieve all available data
        while len(transactions) < response['total_transactions']:
            request = TransactionsGetRequest(
                    access_token=self.access_token,
                    start_date=start_date,
                    end_date=end_date,
                    options=TransactionsGetRequestOptions(
                        offset=len(transactions)
                    )
            )
            response = self.api.transactions_get(request)
            transactions.extend(parse_transaction(response['transactions']))

        # Converting date column from str 
        df = pd.DataFrame(transactions)
        for i in ['date', 'authorized_date']:
            df[i] = pd.to_datetime(df[i]).dt.date

        return df



import textwrap
from pydantic_settings import BaseSettings
from botocore.exceptions import ClientError
from typing import Text, List, Dict, Optional, Any, ClassVar
from pydantic import BaseModel, Field, model_validator, field_validator

from mindsdb.integrations.handlers.bedrock_handler.utilities import create_amazon_bedrock_client
from mindsdb.integrations.utilities.handlers.validation_utilities import ParameterValidationUtilities


class AmazonBedrockHandlerSettings(BaseSettings):
    """
    Settings for Amazon Bedrock handler.

    Attributes
    ----------
    DEFAULT_MODE : Text
        The default mode for the handler.

    SUPPORTED_MODES : List
        List of supported modes for the handler.

    DEFAULT_TEXT_MODEL_ID : Text
        The default model ID to use for text generation. This will be the default model ID for the default and conversational modes.
    """
    # Modes.
    # TODO: Add other modes.
    DEFAULT_MODE: ClassVar[Text] = 'default'
    SUPPORTED_MODES: ClassVar[List] = ['default', 'conversational']

    # TODO: Set the default model ID for other modes.
    # Model IDs.
    DEFAULT_TEXT_MODEL_ID: ClassVar[Text] = 'amazon.titan-text-premier-v1:0'


class AmazonBedrockHandlerEngineConfig(BaseModel):
    """
    Configuration model for engines created via the Amazon Bedrock handler.

    Attributes
    ----------
    aws_access_key_id : Text
        The AWS access key ID.

    aws_secret_access_key : Text
        The AWS secret access key.

    region_name : Text
        The AWS region name.

    aws_session_token : Text, Optional
        The AWS session token. Optional, but required for temporary security credentials.
    """
    aws_access_key_id: Text
    aws_secret_access_key: Text
    region_name: Text
    aws_session_token: Optional[Text] = None

    class Config:
        extra = "forbid"

    @model_validator(mode="before")
    @classmethod
    def check_if_params_contain_typos(cls, values: Any) -> Any:
        """
        Checks if there are any typos in the parameters.

        Args:
            values (Any): The parameters provided when creating an engine via the Amazon Bedrock handler.

        Raises:
            ValueError: If there are any typos in the parameters.
        """
        ParameterValidationUtilities.validate_parameter_spelling(cls, values)

        return values

    @model_validator(mode="after")
    @classmethod
    def check_access_to_amazon_bedrock(cls, model: BaseModel) -> BaseModel:
        """
        Checks if the AWS credentials provided are valid and Amazon Bedrock is accessible.

        Args:
            model (BaseModel): The parameters provided when creating an engine via the Amazon Bedrock handler.

        Raises:
            ValueError: If the AWS credentials are invalid or Amazon Bedrock is not accessible.
        """
        bedrock_client = create_amazon_bedrock_client(
            "bedrock",
            model.aws_access_key_id,
            model.aws_secret_access_key,
            model.region_name,
            model.aws_session_token
        )

        try:
            bedrock_client.list_foundation_models()
        except ClientError as e:
            raise ValueError(f"Invalid Amazon Bedrock credentials: {e}!")

        return model


class AmazonBedrockHandlerModelConfig(BaseModel):
    """
    Configuration model for models created via the Amazon Bedrock handler.

    Attributes
    ----------
    id : Text
        The ID of the model in Amazon Bedrock.

    mode : Optional[Text]
        The mode to run the handler model in. The default mode and the supported modes are defined in the AmazonBedrockHandlerSettings class.

    prompt_template : Optional[Text]
        The base template for prompts with placeholders.

    question_column : Optional[Text]
        The column name for questions to be asked.

    context_column : Optional[Text]
        The column name for context to be provided with the questions.

    temperature : Optional[float]
        The setting for the randomness in the responses generated by the model.

    top_p : Optional[float]
        The setting for the probability of the tokens in the responses generated by the model.

    max_tokens : Optional[int]
        The maximum number of tokens to generate in the responses.

    stop : Optional[List[Text]]
        The list of sequences to stop the generation of tokens in the responses.

    connection_args : Dict
        The connection arguments passed required to connect to Amazon Bedrock. These are AWS credentials provided when creating the engine.
    """
    # User-provided Handler Model Prameters: These are parameters specific to the MindsDB handler for Amazon Bedrock provided by the user.
    id: Text = Field(None)
    mode: Optional[Text] = Field(AmazonBedrockHandlerSettings.DEFAULT_MODE)
    prompt_template: Optional[Text] = Field(None)
    question_column: Optional[Text] = Field(None)
    context_column: Optional[Text] = Field(None)

    # Amazon Bedrock Model Parameters: These are parameters specific to the models in Amazon Bedrock. They are provided by the user.
    temperature: Optional[float] = Field(None, bedrock_model_param=True, bedrock_model_param_name='temperature')
    top_p: Optional[float] = Field(None, bedrock_model_param=True, bedrock_model_param_name='topP')
    max_tokens: Optional[int] = Field(None, bedrock_model_param=True, bedrock_model_param_name='maxTokens')
    stop: Optional[List[Text]] = Field(None, bedrock_model_param=True, bedrock_model_param_name='stopSequences')

    # System-provided Handler Model Parameters: These are parameters specific to the MindsDB handler for Amazon Bedrock provided by the system.
    connection_args: Dict = Field(None, exclude=True)

    class Config:
        extra = "forbid"

    @model_validator(mode="before")
    @classmethod
    def check_if_params_contain_typos(cls, values: Any) -> Any:
        """
        Checks if there are any typos in the parameters.

        Args:
            values (Any): The parameters provided when creating a model via the Amazon Bedrock handler.

        Raises:
            ValueError: If there are any typos in the parameters.
        """
        ParameterValidationUtilities.validate_parameter_spelling(cls, values)

        return values

    @field_validator("mode")
    @classmethod
    def check_if_mode_is_supported(cls, mode: Text) -> Text:
        """
        Checks if the mode provided is supported.

        Args:
            mode (Text): The mode to run the handler model in.

        Raises:
            ValueError: If the mode provided is not supported.
        """
        if mode not in AmazonBedrockHandlerSettings.SUPPORTED_MODES:
            raise ValueError(f"Mode {mode} is not supported. The supported modes are {''.join(AmazonBedrockHandlerSettings.SUPPORTED_MODES)}!")

        return mode

    @model_validator(mode="after")
    @classmethod
    def check_if_model_id_is_valid_and_correct_for_mode(cls, model: BaseModel) -> BaseModel:
        """
        Checks if the model ID and the parameters provided for the model are valid.
        If a model ID is not provided, the default model ID for that mode will be used.

        Args:
            values (Any): The parameters provided when creating a model via the Amazon Bedrock handler.

        Raises:
            ValueError: If the model ID provided is invalid or the parameters provided are invalid for the chosen model.
        """
        # TODO: Set the default model ID for other modes.
        if model.id is None:
            if model.mode in ['default', 'conversational']:
                model.id = AmazonBedrockHandlerSettings.DEFAULT_TEXT_MODEL_ID

        bedrock_client = create_amazon_bedrock_client(
            "bedrock",
            **model.connection_args
        )

        try:
            # Check if the model ID is valid and accessible.
            response = bedrock_client.get_foundation_model(modelIdentifier=model.id)
        except ClientError as e:
            raise ValueError(f"Invalid Amazon Bedrock model ID: {e}!")

        # Check if the model is suitable for the mode provided.
        if model.mode in ['default', 'conversational']:
            if 'TEXT' not in response['modelDetails']['outputModalities']:
                raise ValueError(f"The models used for the {model.mode} should support text generation!")

        return model

    @model_validator(mode="after")
    @classmethod
    def check_if_params_are_valid_for_mode(cls, model: BaseModel) -> BaseModel:
        """
        Checks if the parameters required for the chosen mode provided are valid.

        Args:
            model (BaseModel): The parameters provided when creating a model via the Amazon Bedrock handler.

        Raises:
            ValueError: If the parameters provided are invalid for the mode provided.
        """
        # If the mode is default, one of the following need to be provided:
        # 1. prompt_template.
        # 2. question_column with an optional context_column.
        # TODO: Find the other possible parameters/combinations for the default mode.
        if model.mode in ['default', 'conversational']:
            error_message = textwrap.dedent(
                f"""\
                    For the {model.mode} mode, one of the following need to be provided:
                        1) A `prompt_template`
                        2) A `question_column` and an optional `context_column`
                """
            )
            if model.prompt_template is None and model.question_column is None:
                raise ValueError(error_message)

            if model.prompt_template is not None and model.question_column is not None:
                raise ValueError(error_message)

            if model.context_column is not None and model.question_column is None:
                raise ValueError(error_message)

        # TODO: Add validations for other modes.

        return model

    def model_dump(self) -> Dict:
        """
        Dumps the model configuration to a dictionary.

        Returns:
            Dict: The configuration of the model.
        """
        bedrock_model_param_names = [val.get("bedrock_model_param_name") for key, val in self.model_json_schema(mode='serialization')['properties'].items() if val.get("bedrock_model_param")]
        bedrock_model_params = [key for key, val in self.model_json_schema(mode='serialization')['properties'].items() if val.get("bedrock_model_param")]

        handler_model_params = [key for key, val in self.model_json_schema(mode='serialization')['properties'].items() if not val.get("bedrock_model_param")]

        inference_config = {}
        for index, key in enumerate(bedrock_model_params):
            if getattr(self, key) is not None:
                inference_config[bedrock_model_param_names[index]] = getattr(self, key)

        return {
            "inference_config": inference_config,
            **{key: getattr(self, key) for key in handler_model_params}
        }



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    url={
        'type': ARG_TYPE.STR,
        'description': 'The URL to the Phoenix Query Server.'
    },
    max_retries={
        'type': ARG_TYPE.INT,
        'description': 'The maximum number of retries in case there is a connection error.'
    },
    autocommit={
        'type': ARG_TYPE.BOOL,
        'description': 'The flag for switching the connection to autocommit mode.'
    },
    auth={
        'type': ARG_TYPE.STR,
        'description': 'An authentication configuration object as expected by the underlying python_requests and python_requests_gssapi library.'
    },
    authentication={
        'type': ARG_TYPE.STR,
        'description': 'An alternative way to specify the authentication mechanism that mimics the semantics of the JDBC drirver.'
    },
    avatica_user={
        'type': ARG_TYPE.STR,
        'description': 'The username for BASIC or DIGEST authentication. Use in conjunction with the authentication option.'
    },
    avatica_password={
        'type': ARG_TYPE.PWD,
        'description': 'The password for BASIC or DIGEST authentication. Use in conjunction with the authentication option.',
        'secret': True
    },
    user={
        'type': ARG_TYPE.STR,
        'description': 'If authentication is BASIC or DIGEST then alias for avatica_user. If authentication is NONE or SPNEGO then alias for do_as'
    },
    password={
        'type': ARG_TYPE.PWD,
        'description': 'If authentication is BASIC or DIGEST then alias for avatica_password.',
        'secret': True
    }
)

connection_args_example = OrderedDict(
    url='http://127.0.0.1:8765',
    autocommit=True
)



import unittest
from mindsdb.integrations.handlers.orioledb_handler.orioledb_handler import OrioleDBHandler
from mindsdb.integrations.libs.response import RESPONSE_TYPE

class OrioleDBHandlerTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.kwargs = {
            "connection_data": {
                "host": "localhost",
                "port": 5432,
                "user": "postgres",
                "password": "1234",
                "database": "postgres",
            }
        }
        cls.handler = OrioleDBHandler('test_orioledb_handler', **cls.kwargs)

    def test_0_check_connection(self):
        assert self.handler.check_connection()

    def test_1_connect(self):
        assert self.handler.connect()
    
    def test_2_create_table(self):
        query = "CREATE Table IF NOT EXISTS Lover(name varchar(101));"
        result = self.handler.query(query)
        assert result.type is not RESPONSE_TYPE.ERROR 

    def test_3_insert(self):
        query = "INSERT INTO LOVER VALUES('Shiv Shakti');"
        result = self.handler.query(query)
        assert result.type is not RESPONSE_TYPE.ERROR 

    def test_4_native_query_select(self):
        query = "SELECT * FROM LOVER;"
        result = self.handler.query(query)
        assert result.type is RESPONSE_TYPE.TABLE 

    def test_5_get_tables(self):
        tables = self.handler.get_tables()
        assert tables.type is  RESPONSE_TYPE.TABLE

    def test_6_get_columns(self):
        columns = self.handler.get_columns('LOVER')
        
        query = "DROP Table IF  EXISTS Lover;"
        result = self.handler.query(query)
        assert columns.type is not RESPONSE_TYPE.ERROR


if __name__ == '__main__':
    unittest.main()


    



import datetime as dt

import json

from google_auth_oauthlib.flow import Flow


def credentials_to_dict(credentials):
  return {'token': credentials.token,
          'refresh_token': credentials.refresh_token,
          'token_uri': credentials.token_uri,
          'client_id': credentials.client_id,
          'client_secret': credentials.client_secret,
          'scopes': credentials.scopes,
          'expiry': dt.datetime.strftime(credentials.expiry, '%Y-%m-%dT%H:%M:%S')}


class AuthException(Exception):
    def __init__(self, message, auth_url=None):
        super().__init__(message)

        self.auth_url = auth_url


def google_auth_flow(secret_file, scopes, code=None):
    # initialise flow
    flow = Flow.from_client_secrets_file(secret_file, scopes)

    # get host url from flask
    from flask import request
    flow.redirect_uri = request.headers['ORIGIN'] + '/verify-auth'

    if code:
        flow.fetch_token(code=code)
        creds = flow.credentials
        return creds
    else:
        auth_url = flow.authorization_url()[0]
        raise AuthException(f'Authorisation required. Please follow the url: {auth_url}', auth_url=auth_url)


def save_creds_to_file(creds, file_path):
    with open(file_path, 'w') as token:
        data = credentials_to_dict(creds)
        token.write(json.dumps(data))



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    url={
        "type": ARG_TYPE.URL,
        "description": "Confluence URL",
        "label": "url",
        "required": True
    },
    username={
        "type": ARG_TYPE.STR,
        "description": "Confluence username",
        "label": "username",
        "required": True
    },
    password={
        "type": ARG_TYPE.STR,
        "description": "Password",
        "label": "password",
        "required": True,
        "secret": True
    }
)

connection_args_example = OrderedDict(
    url="https://marios.atlassian.net/",
    username="your_username",
    password="access_token"
)



import copy

from mindsdb_sql_parser.ast.mindsdb import Latest
from mindsdb_sql_parser.ast import (
    Select, Identifier, BetweenOperation, Join, Star, BinaryOperation, Constant,
    OrderBy, NullConstant
)

from mindsdb.integrations.utilities.query_traversal import query_traversal

from mindsdb.api.executor.planner.exceptions import PlanningException
from mindsdb.api.executor.planner import utils
from mindsdb.api.executor.planner.steps import (
    JoinStep, LimitOffsetStep, MultipleSteps, MapReduceStep,
    ApplyTimeseriesPredictorStep
)
from mindsdb.api.executor.planner.ts_utils import validate_ts_where_condition, find_time_filter, replace_time_filter, \
    find_and_remove_time_filter, recursively_check_join_identifiers_for_ambiguity


class PlanJoinTSPredictorQuery:

    def __init__(self, planner):
        self.planner = planner

    def adapt_dbt_query(self, query, integration):
        orig_query = query

        join = query.from_table
        join_left = join.left

        # dbt query.

        # move latest into subquery
        moved_conditions = []

        def move_latest(node, **kwargs):
            if isinstance(node, BinaryOperation):
                if Latest() in node.args:
                    for arg in node.args:
                        if isinstance(arg, Identifier):
                            # remove table alias
                            arg.parts = [arg.parts[-1]]
                    moved_conditions.append(node)

        query_traversal(query.where, move_latest)

        # TODO make project step from query.target

        # TODO support complex query. Only one table is supported at the moment.
        # if not isinstance(join_left.from_table, Identifier):
        #     raise PlanningException(f'Statement not supported: {query.to_string()}')

        # move properties to upper query
        query = join_left

        if query.from_table.alias is not None:
            table_alias = [query.from_table.alias.parts[0]]
        else:
            table_alias = query.from_table.parts

        # add latest to query.where
        for cond in moved_conditions:
            if query.where is not None:
                query.where = BinaryOperation('and', args=[query.where, cond])
            else:
                query.where = cond

        def add_aliases(node, is_table, **kwargs):
            if not is_table and isinstance(node, Identifier):
                if len(node.parts) == 1:
                    # add table alias to field
                    node.parts = table_alias + node.parts

        query_traversal(query.where, add_aliases)

        if isinstance(query.from_table, Identifier):
            # DBT workaround: allow use tables without integration.
            #   if table.part[0] not in integration - take integration name from create table command
            if (
                integration is not None
                and query.from_table.parts[0] not in self.planner.databases
            ):
                # add integration name to table
                query.from_table.parts.insert(0, integration)

        join_left = join_left.from_table

        if orig_query.limit is not None:
            if query.limit is None or query.limit.value > orig_query.limit.value:
                query.limit = orig_query.limit
        query.parentheses = False
        query.alias = None

        return query, join_left

    def get_aliased_fields(self, targets):
        # get aliases from select target
        aliased_fields = {}
        for target in targets:
            if target.alias is not None:
                aliased_fields[target.alias.to_string()] = target
        return aliased_fields

    def plan_fetch_timeseries_partitions(self, query, table, predictor_group_by_names):
        targets = [
            Identifier(column)
            for column in predictor_group_by_names
        ]

        query = Select(
            distinct=True,
            targets=targets,
            from_table=table,
            where=query.where,
            modifiers=query.modifiers,
        )
        select_step = self.planner.plan_integration_select(query)
        return select_step

    def plan(self, query, integration=None):
        # integration is for dbt only

        join = query.from_table
        join_left = join.left
        join_right = join.right

        predictor_is_left = False
        if self.planner.is_predictor(join_left):
            # predictor is in the left, put it in the right
            join_left, join_right = join_right, join_left
            predictor_is_left = True

        if self.planner.is_predictor(join_left):
            # in the left is also predictor
            raise PlanningException(f'Can\'t join two predictors {join_left} and {join_left}')

        orig_query = query
        # dbt query?
        if isinstance(join_left, Select) and isinstance(join_left.from_table, Identifier):
            query, join_left = self.adapt_dbt_query(query, integration)

        predictor_namespace, predictor = self.planner.get_predictor_namespace_and_name_from_identifier(join_right)
        table = join_left

        aliased_fields = self.get_aliased_fields(query.targets)

        recursively_check_join_identifiers_for_ambiguity(query.where)
        recursively_check_join_identifiers_for_ambiguity(query.group_by, aliased_fields=aliased_fields)
        recursively_check_join_identifiers_for_ambiguity(query.having)
        recursively_check_join_identifiers_for_ambiguity(query.order_by, aliased_fields=aliased_fields)

        predictor_steps = self.plan_timeseries_predictor(query, table, predictor_namespace, predictor)

        # add join
        # Update reference

        left = Identifier(predictor_steps['predictor'].result.ref_name)
        right = Identifier(predictor_steps['data'].result.ref_name)

        if not predictor_is_left:
            # swap join
            left, right = right, left
        new_join = Join(left=left, right=right, join_type=join.join_type)

        left = predictor_steps['predictor'].result
        right = predictor_steps['data'].result
        if not predictor_is_left:
            # swap join
            left, right = right, left

        last_step = self.planner.plan.add_step(JoinStep(left=left, right=right, query=new_join))

        # limit from timeseries
        if predictor_steps.get('saved_limit'):
            last_step = self.planner.plan.add_step(
                LimitOffsetStep(dataframe=last_step.result, limit=predictor_steps['saved_limit'])
            )

        return self.planner.plan_project(orig_query, last_step.result)

    def plan_timeseries_predictor(self, query, table, predictor_namespace, predictor):

        predictor_metadata = self.planner.get_predictor(predictor)

        predictor_time_column_name = predictor_metadata['order_by_column']
        predictor_group_by_names = predictor_metadata['group_by_columns']
        if predictor_group_by_names is None:
            predictor_group_by_names = []
        predictor_window = predictor_metadata['window']

        if query.order_by:
            raise PlanningException(
                f'Can\'t provide ORDER BY to time series predictor, it will be taken from predictor settings. Found: {query.order_by}')

        saved_limit = None
        if query.limit is not None:
            saved_limit = query.limit.value

        if query.group_by or query.having or query.offset:
            raise PlanningException(f'Unsupported query to timeseries predictor: {str(query)}')

        allowed_columns = [predictor_time_column_name.lower()]
        if len(predictor_group_by_names) > 0:
            allowed_columns += [i.lower() for i in predictor_group_by_names]

        no_time_filter_query = copy.deepcopy(query)

        preparation_where = no_time_filter_query.where

        validate_ts_where_condition(preparation_where, allowed_columns=allowed_columns)

        time_filter = find_time_filter(preparation_where, time_column_name=predictor_time_column_name)

        order_by = [OrderBy(Identifier(parts=[predictor_time_column_name]), direction='DESC')]

        query_modifiers = query.modifiers

        # add {order_by_field} is not null
        def add_order_not_null(condition):
            order_field_not_null = BinaryOperation(op='is not', args=[
                Identifier(parts=[predictor_time_column_name]),
                NullConstant()
            ])
            if condition is not None:
                condition = BinaryOperation(op='and', args=[
                    condition,
                    order_field_not_null
                ])
            else:
                condition = order_field_not_null
            return condition

        preparation_where2 = copy.deepcopy(preparation_where)
        preparation_where = add_order_not_null(preparation_where)

        # Obtain integration selects
        if isinstance(time_filter, BetweenOperation):
            between_from = time_filter.args[1]
            preparation_time_filter = BinaryOperation('<', args=[Identifier(predictor_time_column_name), between_from])
            preparation_where2 = replace_time_filter(preparation_where2, time_filter, preparation_time_filter)
            integration_select_1 = Select(
                targets=[Star()],
                from_table=table,
                where=add_order_not_null(preparation_where2),
                modifiers=query_modifiers,
                order_by=order_by,
                limit=Constant(predictor_window)
            )

            integration_select_2 = Select(targets=[Star()],
                                          from_table=table,
                                          where=preparation_where,
                                          modifiers=query_modifiers,
                                          order_by=order_by)

            integration_selects = [integration_select_1, integration_select_2]
        elif isinstance(time_filter, BinaryOperation) and time_filter.op == '>' and time_filter.args[1] == Latest():
            integration_select = Select(targets=[Star()],
                                        from_table=table,
                                        where=preparation_where,
                                        modifiers=query_modifiers,
                                        order_by=order_by,
                                        limit=Constant(predictor_window),
                                        )
            integration_select.where = find_and_remove_time_filter(integration_select.where, time_filter)
            integration_selects = [integration_select]
        elif isinstance(time_filter, BinaryOperation) and time_filter.op == '=':
            integration_select = Select(targets=[Star()],
                                        from_table=table,
                                        where=preparation_where,
                                        modifiers=query_modifiers,
                                        order_by=order_by,
                                        limit=Constant(predictor_window),
                                        )

            if type(time_filter.args[1]) is Latest:
                integration_select.where = find_and_remove_time_filter(integration_select.where, time_filter)
            else:
                time_filter_date = time_filter.args[1]
                preparation_time_filter = BinaryOperation(
                    '<=',
                    args=[
                        Identifier(predictor_time_column_name),
                        time_filter_date
                    ]
                )
                integration_select.where = add_order_not_null(
                    replace_time_filter(
                        preparation_where2, time_filter, preparation_time_filter
                    )
                )
                time_filter.op = '>'

            integration_selects = [integration_select]
        elif isinstance(time_filter, BinaryOperation) and time_filter.op in ('>', '>='):
            time_filter_date = time_filter.args[1]
            preparation_time_filter_op = {'>': '<=', '>=': '<'}[time_filter.op]

            preparation_time_filter = BinaryOperation(preparation_time_filter_op, args=[Identifier(predictor_time_column_name), time_filter_date])
            preparation_where2 = replace_time_filter(preparation_where2, time_filter, preparation_time_filter)
            integration_select_1 = Select(targets=[Star()],
                                          from_table=table,
                                          where=add_order_not_null(preparation_where2),
                                          modifiers=query_modifiers,
                                          order_by=order_by,
                                          limit=Constant(predictor_window))

            integration_select_2 = Select(targets=[Star()],
                                          from_table=table,
                                          where=preparation_where,
                                          modifiers=query_modifiers,
                                          order_by=order_by)

            integration_selects = [integration_select_1, integration_select_2]
        else:
            integration_select = Select(targets=[Star()],
                                        from_table=table,
                                        where=preparation_where,
                                        modifiers=query_modifiers,
                                        order_by=order_by,
                                        )
            integration_selects = [integration_select]

        if len(predictor_group_by_names) == 0:
            # ts query without grouping
            # one or multistep
            if len(integration_selects) == 1:
                select_partition_step = self.planner.get_integration_select_step(integration_selects[0])
            else:
                select_partition_step = MultipleSteps(
                    steps=[self.planner.get_integration_select_step(s) for s in integration_selects], reduce='union')

            # fetch data step
            data_step = self.planner.plan.add_step(select_partition_step)
        else:
            # inject $var to queries
            for integration_select in integration_selects:
                condition = integration_select.where
                for num, column in enumerate(predictor_group_by_names):
                    cond = BinaryOperation('=', args=[Identifier(column), Constant(f'$var[{column}]')])

                    # join to main condition
                    if condition is None:
                        condition = cond
                    else:
                        condition = BinaryOperation('and', args=[condition, cond])

                integration_select.where = condition
            # one or multistep
            if len(integration_selects) == 1:
                select_partition_step = self.planner.get_integration_select_step(integration_selects[0])
            else:
                select_partition_step = MultipleSteps(
                    steps=[self.planner.get_integration_select_step(s) for s in integration_selects], reduce='union')

            # get groping values
            no_time_filter_query.where = find_and_remove_time_filter(no_time_filter_query.where, time_filter)
            select_partitions_step = self.plan_fetch_timeseries_partitions(no_time_filter_query, table, predictor_group_by_names)

            # sub-query by every grouping value
            map_reduce_step = self.planner.plan.add_step(MapReduceStep(values=select_partitions_step.result, reduce='union', step=select_partition_step))
            data_step = map_reduce_step

        predictor_identifier = utils.get_predictor_name_identifier(predictor)

        params = None
        if query.using is not None:
            params = query.using
        predictor_step = self.planner.plan.add_step(
            ApplyTimeseriesPredictorStep(
                output_time_filter=time_filter,
                namespace=predictor_namespace,
                dataframe=data_step.result,
                predictor=predictor_identifier,
                params=params,
            )
        )

        return {
            'predictor': predictor_step,
            'data': data_step,
            'saved_limit': saved_limit,
        }



import json
from typing import Optional
from datetime import datetime

from flask import Response


def http_error(status_code: int, title: Optional[str] = None, detail: Optional[str] = None):
    ''' Wrapper for error responce acoording with RFC 7807 (https://tools.ietf.org/html/rfc7807)

        :param status_code: int - http status code for response
        :param title: str
        :param detail: str

        :return: flask Response object
    '''
    if title is None:
        title = 'Error'
    if detail is None:
        if 400 <= status_code < 500:
            detail = "A client error occurred. Please check your request and try again."
        elif 500 <= status_code < 600:
            detail = "A server error occurred. Please try again later."
        else:
            detail = "An error occurred while processing the request. Please try again later."

    return Response(
        response=json.dumps({
            'title': title,
            'detail': detail,
            'timestamp': str(datetime.now())
        }),
        status=status_code,
        headers={
            'Content-Type': 'application/problem+json'
        }
    )



from dataclasses import dataclass
from typing import Callable, List, Union

from langchain_core.documents import Document
from langchain_text_splitters import (
    MarkdownHeaderTextSplitter,
    HTMLHeaderTextSplitter,
    RecursiveCharacterTextSplitter,
)

from mindsdb.interfaces.knowledge_base.preprocessing.models import TextChunkingConfig

from mindsdb.utilities import log

DEFAULT_CHUNK_SIZE = 1000
DEFAULT_CHUNK_OVERLAP = 50
DEFAULT_MARKDOWN_HEADERS_TO_SPLIT_ON = [
    ("#", "Header 1"),
    ("##", "Header 2"),
    ("###", "Header 3"),
]
DEFAULT_HTML_HEADERS_TO_SPLIT_ON = [
    ("h1", "Header 1"),
    ("h2", "Header 2"),
    ("h3", "Header 3"),
    ("h4", "Header 4"),
]
logger = log.getLogger(__name__)


@dataclass
class FileSplitterConfig:
    """Represents configuration needed to split a file into chunks for retrieval."""

    # Target chunk size in characters. Not all splitters will adhere exactly to this (it's more of a guideline)
    chunk_size: int = DEFAULT_CHUNK_SIZE
    # How many characters each chunk should overlap. Not all splitters will adhere exactly to this (it's more of a guideline)
    chunk_overlap: int = DEFAULT_CHUNK_OVERLAP
    # Chunking parameters are passed as a TextChunkingConfig
    text_chunking_config: TextChunkingConfig = None
    # Default recursive splitter to use for text files, or unsupported files
    recursive_splitter: RecursiveCharacterTextSplitter = None
    # Splitter to use for MD splitting
    markdown_splitter: MarkdownHeaderTextSplitter = MarkdownHeaderTextSplitter(
        headers_to_split_on=DEFAULT_MARKDOWN_HEADERS_TO_SPLIT_ON
    )
    # Splitter to use for HTML splitting
    html_splitter: HTMLHeaderTextSplitter = HTMLHeaderTextSplitter(
        headers_to_split_on=DEFAULT_HTML_HEADERS_TO_SPLIT_ON
    )

    def __post_init__(self):
        if self.text_chunking_config is None:
            self.text_chunking_config = TextChunkingConfig(
                chunk_size=self.chunk_size, chunk_overlap=self.chunk_overlap
            )

        if self.recursive_splitter is None:
            self.recursive_splitter = RecursiveCharacterTextSplitter(
                chunk_size=self.text_chunking_config.chunk_size,
                chunk_overlap=self.text_chunking_config.chunk_overlap,
                length_function=self.text_chunking_config.length_function,
                separators=self.text_chunking_config.separators,
            )


class FileSplitter:
    """Splits Documents that represent various file types into chunks for retrieval."""

    def __init__(self, config: FileSplitterConfig):
        """
        Args:
            config (FileSplitterConfig): Configuration for the file splitter.
        """
        self.config = config
        self._extension_map = {
            ".pdf": self._recursive_splitter_fn,
            ".md": self._markdown_splitter_fn,
            ".html": self._html_splitter_fn,
        }
        self.default_splitter = self._recursive_splitter_fn

    def _split_func_by_extension(
        self, extension
    ) -> Union[Callable, HTMLHeaderTextSplitter, MarkdownHeaderTextSplitter]:
        return self._extension_map.get(extension, self.default_splitter)()

    def split_documents(
        self, documents: List[Document], default_failover: bool = True
    ) -> List[Document]:
        """Splits a list of documents representing files using the appropriate splitting & chunking strategies

        Args:
            documents (List[Document]): List of documents representing files to split.
            default_failover (bool, optional): Whether to use the default splitter as a fallback if the file type is not supported. Defaults to True.

        Returns:
            List[Document]: List of documents representing the split files.
        """
        split_documents = []
        document: Document
        for document in documents:
            extension = document.metadata.get("extension")
            split_func = self._split_func_by_extension(extension=extension)
            try:
                split_documents += split_func(document.page_content)
            except Exception as e:
                logger.error(
                    f"Error splitting document with extension {extension}: {str(e)}"
                )
                if not default_failover:
                    raise ValueError(
                        f"Error splitting document with extension {extension}"
                    ) from e
                # Try default splitter as a failover, if enabled.
                split_func = self._split_func_by_extension(extension=None)
                split_documents += split_func(document.page_content)
        return split_documents

    def _markdown_splitter_fn(self) -> Callable:
        return self.config.markdown_splitter.split_text

    def _html_splitter_fn(self) -> Callable:
        return self.config.html_splitter.split_text

    def _recursive_splitter_fn(self) -> Callable:
        # Recursive splitter is a TextSplitter where split_text returns List[str].
        def recursive_split(content: str) -> List[Document]:
            split_content = self.config.recursive_splitter.split_text(content)
            return [Document(page_content=c) for c in split_content]

        return recursive_split



import unittest

from mindsdb.integrations.handlers.hive_handler.hive_handler import HiveHandler
from mindsdb.api.executor.data_types.response_type import RESPONSE_TYPE


class HiveHandlerTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.kwargs = {
            "host": "localhost",
            "port": "10000",
            "user": "admin",
            "password": "password",
            "database": "default",
            "auth": "CUSTOM"
        }
        cls.handler = HiveHandler('test_hive_handler', **cls.kwargs)

    def test_0_connect(self):
        self.handler.check_connection()

    def test_1_native_query_show_dbs(self):
        dbs = self.handler.native_query("SHOW DATABASES;")
        assert dbs['type'] is not RESPONSE_TYPE.ERROR

    def test_2_get_tables(self):
        tbls = self.handler.get_tables()
        assert tbls['type'] is not RESPONSE_TYPE.ERROR

    def test_5_drop_table(self):
        res = self.handler.native_query("DROP TABLE IF EXISTS test_hdb")
        assert res['type'] is not RESPONSE_TYPE.ERROR

    def test_4_create_table(self):
        res = self.handler.native_query("CREATE TABLE IF NOT EXISTS test_hdb (test_col INT)")
        assert res['type'] is not RESPONSE_TYPE.ERROR

    def test_6_describe_table(self):
        described = self.handler.get_columns("test_hdb")
        assert described['type'] is RESPONSE_TYPE.TABLE

    def test_7_select_query(self):
        query = "SELECT * FROM test_mdb WHERE foo=238"
        result = self.handler.query(query)
        assert result['type'] is RESPONSE_TYPE.TABLE


if __name__ == '__main__':
    unittest.main()


import difflib


class ParameterValidationUtilities:
    @staticmethod
    def validate_parameter_spelling(handler_cls, parameters):
        expected_params = handler_cls.model_fields.keys()
        for key in parameters.keys():
            if key not in expected_params:
                close_matches = difflib.get_close_matches(
                    key, expected_params, cutoff=0.4
                )
                if close_matches:
                    raise ValueError(
                        f"Unexpected parameter '{key}'. Did you mean '{close_matches[0]}'?"
                    )
                else:
                    raise ValueError(f"Unexpected parameter '{key}'.")



import time
from collections import UserDict


class HandlersCache(UserDict):
    def __init__(self, max_size: int = 5) -> None:
        self._max_size = max_size
        super().__init__()

    def __setitem__(self, key, value) -> None:
        if len(self.data) > self._max_size:
            sorted_elements = sorted(
                self.data.items(),
                key=lambda x: x[1]['last_usage_at']
            )
            del self.data[sorted_elements[0][0]]
        self.data[key] = {
            'last_usage_at': time.time(),
            'handler': value
        }

    def __getitem__(self, key: int) -> object:
        el = super().__getitem__(key)
        el['last_usage_at'] = time.time()
        return el['handler']


handlers_cacher = HandlersCache()



from copy import deepcopy
import datetime as dt
import os
import threading
from typing import Any, Callable, Dict, List, Text

import pandas as pd
from slack_sdk import WebClient
from slack_sdk.errors import SlackApiError
from slack_sdk.socket_mode import SocketModeClient
from slack_sdk.socket_mode.response import SocketModeResponse
from slack_sdk.socket_mode.request import SocketModeRequest

from mindsdb.integrations.handlers.slack_handler.slack_tables import (
    SlackConversationsTable,
    SlackMessagesTable,
    SlackThreadsTable,
    SlackUsersTable
)
from mindsdb.integrations.libs.api_handler import APIChatHandler, FuncParser
from mindsdb.integrations.libs.response import (
    HandlerResponse as Response,
    HandlerStatusResponse as StatusResponse,
    RESPONSE_TYPE
)
from mindsdb.utilities.config import Config
from mindsdb.utilities import log

logger = log.getLogger(__name__)


class SlackHandler(APIChatHandler):
    """
    This handler handles the connection and execution of SQL statements on Slack.
    Additionally, it allows the setup of a real-time connection to the Slack API using the Socket Mode for chat-bots.
    """

    def __init__(self, name: Text, connection_data: Dict, **kwargs: Any) -> None:
        """
        Initializes the handler.

        Args:
            name (Text): The name of the handler instance.
            connection_data (Dict): The connection data required to connect to the SAP HANA database.
            kwargs(Any): Arbitrary keyword arguments.
        """
        super().__init__(name)
        self.connection_data = connection_data
        self.kwargs = kwargs

        # If the parameters are not provided, check the environment variables and the handler configuration.
        handler_config = Config().get('slack_handler', {})

        for key in ['token', 'app_token']:
            if key not in self.connection_data:
                if f'SLACK_{key.upper()}' in os.environ:
                    self.connection_data[key] = os.environ[f'SLACK_{key.upper()}']
                elif key in handler_config:
                    self.connection_data[key] = handler_config[key]

        self.web_connection = None
        self._socket_connection = None
        self.is_connected = False

        self._register_table('conversations', SlackConversationsTable(self))
        self._register_table('messages', SlackMessagesTable(self))
        self._register_table('threads', SlackThreadsTable(self))
        self._register_table('users', SlackUsersTable(self))

    def connect(self) -> WebClient:
        """
        Establishes a connection to the Slack API using the WebClient.

        Returns:
            WebClient: The WebClient object to interact with the Slack API.
        """
        if self.is_connected is True:
            return self.web_connection

        # Check if the mandatory connection parameter (token) is available.
        if 'token' not in self.connection_data:
            raise ValueError('Required parameter (token) must be provided.')

        try:
            self.web_connection = WebClient(token=self.connection_data['token'])
            self.is_connected = True
            return self.web_connection
        except Exception as unknown_error:
            logger.error(f'Unknown error connecting to Slack API: {unknown_error}')
            raise

    def check_connection(self) -> StatusResponse:
        """
        Checks the status of the connection to the Slack API, both for the WebClient and the Socket Mode.

        Raises:
            SlackApiError: If an error occurs while connecting to the Slack API.

        Returns:
            StatusResponse: An object containing the success status and an error message if an error occurs.
        """
        response = StatusResponse(False)

        try:
            web_connection = self.connect()
            # Check the status of the web connection.
            web_connection.auth_test()

            # Check the status of the socket connection if the app_token is provided.
            if 'app_token' in self.connection_data:
                _socket_connection = SocketModeClient(
                    app_token=self.connection_data['app_token'],
                    web_client=web_connection
                )
                _socket_connection.connect()
                _socket_connection.disconnect()

            response.success = True
        except (SlackApiError, ValueError) as known_error:
            logger.error(f'Connection check to the Slack API failed, {known_error}!')
            response.error_message = str(known_error)
        except Exception as unknown_error:
            logger.error(f'Connection check to the Slack API failed due to an unknown error, {unknown_error}!')
            response.error_message = str(unknown_error)

        if response.success is False and self.is_connected is True:
            self.is_connected = False

        return response

    def native_query(self, query: Text = None) -> Response:
        """
        Executes native Slack SDK methods as specified in the query string.

        Args:
            query: The query string containing the method name and parameters.

        Returns:
            Response: A response object containing the result of the query.
        """
        method_name, params = FuncParser().from_string(query)

        df = self._call_slack_api(method_name, params)

        return Response(
            RESPONSE_TYPE.TABLE,
            data_frame=df
        )

    def _call_slack_api(self, method_name: Text = None, params: Dict = None) -> List[Dict]:
        """
        Calls the Slack SDK method with the specified method name and parameters.

        Args:
            method_name (Text): The name of the method to call.
            params (Dict): The parameters to pass to the method.

        Raises:
            SlackApiError: If an error occurs while calling the Slack SDK method

        Returns:
            List[Dict]: The result from running the Slack SDK method.
        """
        web_connection = self.connect()
        method = getattr(web_connection, method_name)

        items = []
        try:
            response = method(**params)
            response_data = deepcopy(response.data)

            # Get only the data items from the response data.
            items.extend(self._extract_data_from_response(response_data))

            # If the response contains a cursor, fetch the next page of results.
            if 'response_metadata' in response and 'next_cursor' in response['response_metadata']:
                while response['response_metadata']['next_cursor']:
                    response = method(cursor=response['response_metadata']['next_cursor'], **params)
                    response_data = deepcopy(response.data)

                    # Get only the data items from the response data.
                    items.extend(self._extract_data_from_response(response_data))
        except SlackApiError as slack_error:
            error = f"Error calling method '{method_name}' with params '{params}': {slack_error.response['error']}"
            logger.error(error)
            raise

        if items:
            df = pd.DataFrame(items)

        return df

    def _extract_data_from_response(self, response_data: Dict) -> List[Dict]:
        """
        Extracts the data items from the response object.

        Args:
            response_data (Dict): The response object containing the data items.

        Raises:
            ValueError: If the response data could not be parsed.

        Returns:
            List[Dict]: The data items extracted from the response object.
        """
        # Remove the metadata from the response.
        for key in ['ok', 'response_metadata', 'cache_ts', 'latest', 'pin_count', 'has_more']:
            if key in response_data:
                response_data.pop(key)

        # If the response contains only one key, return the value of that key as a list.
        if len(response_data) == 1:
            key = list(response_data.keys())[0]
            if isinstance(response_data[key], list):
                return response_data[key]

            else:
                return [response_data[key]]

        # Otherwise, raise an error.
        raise ValueError('Response data could not be parsed.')

    def get_chat_config(self) -> Dict:
        """
        Returns the chat configuration for the Slack handler.

        Returns:
            Dict: The chat configuration.
        """
        return {
            'polling': {
                'type': 'realtime',
            },
            'tables': [
                {
                    'chat_table': {
                        'name': 'messages',
                        'chat_id_col': 'channel_id',
                        'username_col': 'user',
                        'text_col': 'text',
                        'time_col': 'thread_ts',
                    }
                },
                {
                    'chat_table': {
                        'name': 'threads',
                        'chat_id_col': ['channel_id', 'thread_ts'],
                        'username_col': 'user',
                        'text_col': 'text',
                        'time_col': 'thread_ts',
                    }
                }
            ]
        }

    def get_my_user_name(self) -> Text:
        """
        Gets the name of the bot user.

        Returns:
            Text: The name of the bot user.
        """
        web_connection = self.connect()
        user_info = web_connection.auth_test().data
        return user_info['bot_id']

    def subscribe(self, stop_event: threading.Event, callback: Callable, **kwargs: Any) -> None:
        """
        Subscribes to the Slack API using the Socket Mode for real-time responses to messages.

        Args:
            stop_event (threading.Event): The event to stop the subscription.
            callback (Callable): The callback function to process the messages.
            table_name (Text): The name of the table to subscribe to.
            kwargs: Arbitrary keyword arguments.
        """
        self._socket_connection = SocketModeClient(
            # This app-level token will be used only for establishing a connection.
            app_token=self.connection_data['app_token'],  # xapp-A111-222-xyz
            # The WebClient for performing Web API calls in listeners.
            web_client=WebClient(token=self.connection_data['token']),  # xoxb-111-222-xyz
        )

        def _process_websocket_message(client: SocketModeClient, request: SocketModeRequest) -> None:
            """
            Pre-processes the incoming WebSocket message from the Slack API and calls the callback function to process the message.

            Args:
                client (SocketModeClient): The client object to send the response.
                request (SocketModeRequest): The request object containing the payload.
            """
            # Acknowledge the request.
            response = SocketModeResponse(envelope_id=request.envelope_id)
            client.send_socket_mode_response(response)

            # Ignore requests that are not events.
            if request.type != 'events_api':
                return

            # Ignore duplicate requests.
            if request.retry_attempt is not None and request.retry_attempt > 0:
                return

            payload_event = request.payload['event']
            # Avoid responding to events other than direct messages and app mentions.
            if payload_event['type'] not in ('message', 'app_mention'):
                return

            # Avoid responding to unrelated events like message_changed, message_deleted, etc.
            if 'subtype' in payload_event:
                return

            # Avoid responding to messages from the bot.
            if 'bot_id' in payload_event:
                return

            key = {
                'channel_id': payload_event['channel'],
            }

            row = {
                'text': payload_event['text'],
                'user': payload_event['user'],
                'created_at': dt.datetime.fromtimestamp(float(payload_event['ts'])).strftime('%Y-%m-%d %H:%M:%S')
            }

            # Add thread_ts to the key and row if it is a thread message. This is used to identify threads.
            # This message should be handled via the threads table.
            if 'thread_ts' in payload_event:
                key['thread_ts'] = payload_event['thread_ts']

            callback(row, key)

        self._socket_connection.socket_mode_request_listeners.append(_process_websocket_message)
        self._socket_connection.connect()

        stop_event.wait()

        self._socket_connection.close()



from typing import Iterable, List, Optional

import re
from mindsdb_sql_parser.ast import Select, Show, Describe, Explain

import pandas as pd
from mindsdb_sql_parser import parse_sql
from mindsdb_sql_parser.ast import Identifier
from mindsdb.integrations.utilities.query_traversal import query_traversal

from mindsdb.utilities import log
from mindsdb.utilities.context import context as ctx

logger = log.getLogger(__name__)


class SQLAgent:

    def __init__(
            self,
            command_executor,
            database: str,
            include_tables: Optional[List[str]] = None,
            ignore_tables: Optional[List[str]] = None,
            sample_rows_in_table_info: int = 3,
            cache: Optional[dict] = None
    ):
        self._command_executor = command_executor

        self._sample_rows_in_table_info = int(sample_rows_in_table_info)

        self._tables_to_include = include_tables
        self._tables_to_ignore = []
        self._databases = database.split(',')
        if not self._tables_to_include:
            # ignore_tables and include_tables should not be used together.
            # include_tables takes priority if it's set.
            self._tables_to_ignore = ignore_tables or []
        self._cache = cache

    def _call_engine(self, query: str, database=None):
        # switch database

        ast_query = parse_sql(query.strip('`'))
        self._check_permissions(ast_query)

        if database is None:
            # if we use tables with prefixes it should work for any database
            database = self._databases[0]

        ret = self._command_executor.execute_command(
            ast_query,
            database_name=database
        )
        return ret

    def _check_permissions(self, ast_query):

        # check type of query
        if not isinstance(ast_query, (Select, Show, Describe, Explain)):
            raise ValueError(f"Query is not allowed: {ast_query.to_string()}")

        # Check tables
        if self._tables_to_include:
            def _check_f(node, is_table=None, **kwargs):
                if is_table and isinstance(node, Identifier):
                    name1 = node.to_string()
                    name2 = '.'.join(node.parts)
                    name3 = node.parts[-1]
                    if not {name1, name2, name3}.intersection(self._tables_to_include):
                        raise ValueError(f"Table {name1} not found. Available tables: {', '.join(self._tables_to_include)}")

            query_traversal(ast_query, _check_f)

    def get_usable_table_names(self) -> Iterable[str]:

        cache_key = f'{ctx.company_id}_{",".join(self._databases)}_tables'

        # first check cache and return if found
        if self._cache:
            cached_tables = self._cache.get(cache_key)
            if cached_tables:
                return cached_tables

        if self._tables_to_include:
            return self._tables_to_include

        ret = self._call_engine('show databases;')
        dbs = [lst[0] for lst in ret.data.to_lists() if lst[0] != 'information_schema']
        usable_tables = []
        for db in dbs:
            if db != 'mindsdb' and db in self._databases:
                try:
                    ret = self._call_engine('show tables', database=db)
                    tables = [lst[0] for lst in ret.data.to_lists() if lst[0] != 'information_schema']
                    for table in tables:
                        # By default, include all tables in a database unless expilcitly ignored.
                        table_name = f'{db}.{table}'
                        if table_name not in self._tables_to_ignore:
                            usable_tables.append(table_name)
                except Exception as e:
                    logger.warning('Unable to get tables for %s: %s', db, str(e))
        if self._cache:
            self._cache.set(cache_key, set(usable_tables))

        return usable_tables

    def _resolve_table_names(self, table_names: List[str], all_tables: List[Identifier]) -> List[Identifier]:
        """
        Tries to find table (which comes directly from an LLM) by its name
        Handles backticks (`) and tables without databases
        """

        # index to lookup table
        tables_idx = {}
        for table in all_tables:
            # by name
            tables_idx[(table.parts[-1],)] = table
            # by path
            tables_idx[tuple(table.parts)] = table

        tables = []
        for table_name in table_names:
            if not table_name.strip():
                continue

            # Some LLMs (e.g. gpt-4o) may include backticks or quotes when invoking tools.
            table_name = table_name.strip(' `"\'\n\r')
            table = Identifier(table_name)

            # resolved table
            table2 = tables_idx.get(tuple(table.parts))

            if table2 is None:
                raise ValueError(f"Table {table} not found in database")
            tables.append(table2)

        return tables

    def get_table_info(self, table_names: Optional[List[str]] = None) -> str:
        """ Get information about specified tables.
        Follows best practices as specified in: Rajkumar et al, 2022 (https://arxiv.org/abs/2204.00498)
        If `sample_rows_in_table_info`, the specified number of sample rows will be
        appended to each table description. This can increase performance as demonstrated in the paper.
        """

        all_tables = [Identifier(name) for name in self.get_usable_table_names()]

        if table_names is not None:
            all_tables = self._resolve_table_names(table_names, all_tables)

        tables_info = []
        for table in all_tables:
            key = f"{ctx.company_id}_{table}_info"
            table_info = self._cache.get(key) if self._cache else None
            if table_info is None:
                table_info = self._get_single_table_info(table)
                if self._cache:
                    self._cache.set(key, table_info)

            tables_info.append(table_info)

        return "\n\n".join(tables_info)

    def _get_single_table_info(self, table: Identifier) -> str:
        if len(table.parts) < 2:
            raise ValueError(f"Database is required for table: {table}")
        integration, table_name = table.parts[-2:]
        table_str = str(table)

        dn = self._command_executor.session.datahub.get(integration)

        fields, dtypes = [], []
        for column in dn.get_table_columns(table_name):
            fields.append(column['name'])
            dtypes.append(column.get('type', ''))

        info = f'Table named `{table_name}`\n'
        info += f"\n/* Sample with first {self._sample_rows_in_table_info} rows from table {table_str}:\n"
        info += "\t".join([field for field in fields])
        info += self._get_sample_rows(table_str, fields) + "\n*/"
        info += '\nColumn data types: ' + ",\t".join(
            [f'`{field}` : `{dtype}`' for field, dtype in zip(fields, dtypes)]) + '\n'  # noqa
        return info

    def _get_sample_rows(self, table: str, fields: List[str]) -> str:
        command = f"select {','.join(fields)} from {table} limit {self._sample_rows_in_table_info};"
        try:
            ret = self._call_engine(command)
            sample_rows = ret.data.to_lists()
            sample_rows = list(
                map(lambda ls: [str(i) if len(str(i)) < 100 else str[:100] + '...' for i in ls], sample_rows))
            sample_rows_str = "\n" + "\n".join(["\t".join(row) for row in sample_rows])
        except Exception as e:
            logger.warning(e)
            sample_rows_str = "\n" + "\t [error] Couldn't retrieve sample rows!"

        return sample_rows_str

    def _clean_query(self, query: str) -> str:
        # Sometimes LLM can input markdown into query tools.
        cmd = re.sub(r'```(sql)?', '', query)
        return cmd

    def query(self, command: str, fetch: str = "all") -> str:
        """Execute a SQL command and return a string representing the results.
        If the statement returns rows, a string of the results is returned.
        If the statement returns no rows, an empty string is returned.
        """

        def _tidy(result: List) -> str:
            return '\n'.join(['\t'.join([str(value) for value in row]) for row in result])

        def _repr_result(ret):
            limit_rows = 30

            columns_str = ', '.join([repr(col.name) for col in ret.columns])
            res = f'Output columns: {columns_str}\n'

            data = ret.to_lists()
            if len(data) > limit_rows:
                df = pd.DataFrame(data, columns=[col.name for col in ret.columns])

                res += f'Result has {len(data)} rows. Description of data:\n'
                res += str(df.describe(include='all')) + '\n\n'
                res += f'First {limit_rows} rows:\n'

            else:
                res += 'Result:\n'

            res += _tidy(data[:limit_rows])
            return res

        ret = self._call_engine(self._clean_query(command))
        if fetch == "all":
            result = _repr_result(ret.data)
        elif fetch == "one":
            result = _tidy(ret.data.to_lists()[0])
        else:
            raise ValueError("Fetch parameter must be either 'one' or 'all'")
        return str(result)

    def get_table_info_safe(self, table_names: Optional[List[str]] = None) -> str:
        try:
            return self.get_table_info(table_names)
        except Exception as e:
            return f"Error: {e}"

    def query_safe(self, command: str, fetch: str = "all") -> str:
        try:
            return self.query(command, fetch)
        except Exception as e:
            msg = f"Error: {e}"
            if 'does not exist' in msg and ' relation ' in msg:
                msg += '\nAvailable tables: ' + ', '.join(self.get_usable_table_names())
            return msg



import hashlib

from mindsdb.api.executor.planner.steps import UnionStep

from mindsdb.api.executor.sql_query.result_set import ResultSet
from mindsdb.api.executor.exceptions import WrongArgumentError

from .base import BaseStepCall


class UnionStepCall(BaseStepCall):

    bind = UnionStep

    def call(self, step):
        left_result = self.steps_data[step.left.step_num]
        right_result = self.steps_data[step.right.step_num]

        # count of columns have to match
        if len(left_result.columns) != len(right_result.columns):
            raise WrongArgumentError(
                f'UNION columns count mismatch: {len(left_result.columns)} != {len(right_result.columns)} ')

        # types have to match
        # TODO: return checking type later
        # for i, left_col in enumerate(left_result.columns):
        #     right_col = right_result.columns[i]
        #     type1, type2 = left_col.type, right_col.type
        #     if type1 is not None and type2 is not None:
        #         if type1 != type2:
        #             raise ErSqlWrongArguments(f'UNION types mismatch: {type1} != {type2}')

        data = ResultSet()
        for col in left_result.columns:
            data.add_column(col)

        records_hashes = []
        results = []
        for row in left_result.to_lists():
            if step.unique:
                checksum = hashlib.sha256(str(row).encode()).hexdigest()
                if checksum in records_hashes:
                    continue
                records_hashes.append(checksum)
            results.append(list(row))

        for row in right_result.to_lists():
            if step.unique:
                checksum = hashlib.sha256(str(row).encode()).hexdigest()
                if checksum in records_hashes:
                    continue
                records_hashes.append(checksum)
            results.append(list(row))
        data.add_raw_values(results)

        return data



from mindsdb.integrations.handlers.strapi_handler.strapi_tables import StrapiTable
from mindsdb.integrations.libs.api_handler import APIHandler
from mindsdb.integrations.libs.response import HandlerStatusResponse as StatusResponse
from mindsdb_sql_parser import parse_sql
from mindsdb.utilities import log
import requests
from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE
from collections import OrderedDict
import pandas as pd

logger = log.getLogger(__name__)


class StrapiHandler(APIHandler):
    def __init__(self, name: str, **kwargs) -> None:
        """initializer method

        Args:
            name (str): handler name
        """
        super().__init__(name)

        self.connection = None
        self.is_connected = False
        args = kwargs.get('connection_data', {})
        if 'host' in args and 'port' in args:
            self._base_url = f"http://{args['host']}:{args['port']}"
        if 'api_token' in args:
            self._api_token = args['api_token']
        if 'plural_api_ids' in args:
            self._plural_api_ids = args['plural_api_ids']
        # Registers tables for each collections in strapi
        for pluralApiId in self._plural_api_ids:
            self._register_table(table_name=pluralApiId, table_class=StrapiTable(handler=self, name=pluralApiId))

    def check_connection(self) -> StatusResponse:
        """checking the connection

        Returns:
            StatusResponse: whether the connection is still up
        """
        response = StatusResponse(False)
        try:
            self.connect()
            response.success = True
        except Exception as e:
            logger.error(f'Error connecting to Strapi API: {e}!')
            response.error_message = e

        self.is_connected = response.success
        return response

    def connect(self) -> StatusResponse:
        """making the connectino object
        """
        if self.is_connected and self.connection:
            return self.connection

        try:
            headers = {"Authorization": f"Bearer {self._api_token}"}
            response = requests.get(f"{self._base_url}", headers=headers)
            if response.status_code == 200:
                self.connection = response
                self.is_connected = True
                return StatusResponse(True)
            else:
                raise Exception(f"Error connecting to Strapi API: {response.status_code} - {response.text}")
        except Exception as e:
            logger.error(f'Error connecting to Strapi API: {e}!')
            return StatusResponse(False, error_message=e)

    def native_query(self, query: str) -> StatusResponse:
        """Receive and process a raw query.

        Parameters
        ----------
        query : str
            query in a native format

        Returns
        -------
        StatusResponse
            Request status
        """
        ast = parse_sql(query)
        return self.query(ast)

    def call_strapi_api(self, method: str, endpoint: str, params: dict = {}, json_data: dict = {}) -> pd.DataFrame:
        headers = {"Authorization": f"Bearer {self._api_token}"}
        url = f"{self._base_url}{endpoint}"

        if method.upper() in ('GET', 'POST', 'PUT', 'DELETE'):
            headers['Content-Type'] = 'application/json'

            if method.upper() in ('POST', 'PUT', 'DELETE'):
                response = requests.request(method, url, headers=headers, params=params, data=json_data)
            else:
                response = requests.get(url, headers=headers, params=params)

            if response.status_code == 200:
                data = response.json()
                # Create an empty DataFrame
                df = pd.DataFrame()
                if isinstance(data.get('data', None), list):
                    for item in data['data']:
                        # Add 'id' and 'attributes' to the DataFrame
                        row_data = {'id': item['id'], **item['attributes']}
                        df = df._append(row_data, ignore_index=True)
                    return df
                elif isinstance(data.get('data', None), dict):
                    # Add 'id' and 'attributes' to the DataFrame
                    row_data = {'id': data['data']['id'], **data['data']['attributes']}
                    df = df._append(row_data, ignore_index=True)
                    return df
            else:
                raise Exception(f"Error connecting to Strapi API: {response.status_code} - {response.text}")

        return pd.DataFrame()


connection_args = OrderedDict(
    api_token={
        "type": ARG_TYPE.PWD,
        "description": "Strapi API key to use for authentication.",
        "required": True,
        "label": "Api token",
    },
    host={
        "type": ARG_TYPE.URL,
        "description": "Strapi API host to connect to.",
        "required": True,
        "label": "Host",
    },
    port={
        "type": ARG_TYPE.INT,
        "description": "Strapi API port to connect to.",
        "required": True,
        "label": "Port",
    },
    plural_api_ids={
        "type": list,
        "description": "Plural API id to use for querying.",
        "required": True,
        "label": "Plural API id",
    },
)

connection_args_example = OrderedDict(
    host="localhost",
    port=1337,
    api_token="c56c000d867e95848c",
    plural_api_ids=["posts", "portfolios"],
)



from .__about__ import (
    __description__ as description,
    __version__ as version,
)
from mindsdb.integrations.libs.const import HANDLER_TYPE

try:
    from .popularity_recommender_handler import PopularityRecommenderHandler as Handler

    import_error = None
except Exception as e:
    Handler = None
    import_error = e

title = "Popularity Recommender"
name = "popularity_recommender"
type = HANDLER_TYPE.ML
icon_path = "icon.svg"
permanent = False

__all__ = ["Handler", "version", "name", "type", "title", "description", "import_error", "icon_path"]



from typing import Optional, Dict
import pandas as pd

from mindsdb.integrations.handlers.stabilityai_handler.stabilityai import StabilityAPIClient

from mindsdb.integrations.libs.base import BaseMLEngine

from mindsdb.utilities import log

from mindsdb.integrations.utilities.handler_utils import get_api_key


logger = log.getLogger(__name__)


class StabilityAIHandler(BaseMLEngine):
    name = "stabilityai"

    @staticmethod
    def create_validation(target, args=None, **kwargs):
        args = args['using']

        available_tasks = ["text-to-image", "image-to-image", "image-upscaling", "image-masking"]

        if 'task' not in args:
            raise Exception('task has to be specified. Available tasks are - ' + available_tasks)

        if args['task'] not in available_tasks:
            raise Exception('Unknown task specified. Available tasks are - ' + available_tasks)

        if 'local_directory_path' not in args:
            raise Exception('local_directory_path has to be specified')

        client = StabilityAPIClient(args["stabilityai_api_key"], args["local_directory_path"])

        if "engine_id" in args:
            if not client._is_valid_engine(args["engine_id"]):
                raise Exception("Unknown engine. The available engines are - " + list(client.available_engines.keys()))
        else:
            args["engine_id"] = "stable-diffusion-xl-1024-v1-0"

        if "upscale_engine_id" in args:
            if not client._is_valid_engine(args["upscale_engine_id"]):
                raise Exception("Unknown engine. The available engines are - " + list(client.available_engines.keys()))
        else:
            args["upscale_engine_id"] = "esrgan-v1-x2plus"

    def create(self, target: str, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None) -> None:
        if 'using' not in args:
            raise Exception("Stability AI Inference engine requires a USING clause! Refer to its documentation for more details.")
        self.generative = True

        args = args['using']
        args['target'] = target
        self.model_storage.json_set('args', args)

    def _get_stability_client(self, args):
        api_key = get_api_key('stabilityai', args["using"], self.engine_storage, strict=False)

        local_directory_path = args["local_directory_path"]
        engine_id = args.get('engine_id', "stable-diffusion-xl-1024-v1-0")

        return StabilityAPIClient(api_key=api_key, dir_to_save=local_directory_path, engine=engine_id)

    def _process_text_image(self, df, args):

        def generate_text_image(conds, client):
            conds = conds.to_dict()
            return client.text_to_image(prompt=conds.get("text"), height=conds.get("height", 1024), width=conds.get("width", 1024))

        supported_params = set(["text", "height", "width"])

        if "text" not in df.columns:
            raise Exception("`text` column has to be given in the query.")

        for col in df.columns:
            if col not in supported_params:
                raise Exception(f"Unknown column {col}. Currently supported parameters for text to image - {supported_params}")

        client = self._get_stability_client(args)

        return df[df.columns.intersection(supported_params)].apply(generate_text_image, client=client, axis=1)

    def _process_image_image(self, df, args):

        def generate_image_image(conds, client):
            conds = conds.to_dict()
            return client.image_to_image(image_url=conds.get("image_url"), prompt=conds.get("text"), height=conds.get("height", 1024), width=conds.get("width", 1024))

        supported_params = set(["image_url", "text", "height", "width"])

        if "image_url" not in df.columns:
            raise Exception("`image_url` column has to be given in the query.")

        for col in df.columns:
            if col not in supported_params:
                raise Exception(f"Unknown column {col}. Currently supported parameters for image to image - {supported_params}")

        client = self._get_stability_client(args)

        return df[df.columns.intersection(supported_params)].apply(generate_image_image, client=client, axis=1)

    def _process_image_upscaling(self, df, args):

        def generate_image_upscaling(conds, client):
            conds = conds.to_dict()
            return client.image_upscaling(image_url=conds.get("image_url"), prompt=conds.get("text"), height=conds.get("height"), width=conds.get("width"))

        supported_params = set(["image_url", "text", "height", "width"])

        if "image_url" not in df.columns:
            raise Exception("`image_url` column has to be given in the query.")

        for col in df.columns:
            if col not in supported_params:
                raise Exception(f"Unknown column {col}. Currently supported parameters for image scaling - {supported_params}")

        client = self._get_stability_client(args)

        return df[df.columns.intersection(supported_params)].apply(generate_image_upscaling, client=client, axis=1)

    def _process_image_masking(self, df, args):

        def generate_image_mask(conds, client):
            conds = conds.to_dict()
            return client.image_to_image(image_url=conds.get("image_url"), prompt=conds.get("text"), height=conds.get("height"), width=conds.get("width"), mask_image_url=conds.get("mask_image_url"))

        supported_params = set(["image_url", "text", "height", "width", "mask_image_url"])

        if "image_url" not in df.columns:
            raise Exception("`image_url` column has to be given in the query.")

        if "mask_image_url" not in df.columns:
            raise Exception("`mask_image_url` column has to be given in the query.")

        for col in df.columns:
            if col not in supported_params:
                raise Exception(f"Unknown column {col}. Currently supported parameters for image masking - {supported_params}")

        client = self._get_stability_client(args)

        return df[df.columns.intersection(supported_params)].apply(generate_image_mask, client=client, axis=1)

    def predict(self, df, args=None):

        args = self.model_storage.json_get('args')

        if args["task"] == "text-to-image":
            preds = self._process_text_image(df, args)
        elif args["task"] == "image-to-image":
            preds = self._process_image_image(df, args)
        elif args["task"] == "image-upscaling":
            preds = self._process_image_upscaling(df, args)
        elif args["task"] == "image-masking":
            preds = self._process_image_masking(df, args)

        result_df = pd.DataFrame()

        result_df['predictions'] = preds

        result_df = result_df.rename(columns={'predictions': args['target']})

        return result_df

    def describe(self, attribute: Optional[str] = None) -> pd.DataFrame:
        args = self.model_storage.json_get('args')
        client = StabilityAPIClient(args["stabilityai_api_key"], "")
        engine_id = args["engine_id"]
        upscale_engine_id = args["upscale_engine_id"]
        engine_id_res = client.available_engines.get(engine_id)
        upscale_engine_id_res = client.available_engines.get(upscale_engine_id)
        return pd.json_normalize([engine_id_res, upscale_engine_id_res])



from io import BytesIO
from typing import List, Text

import pandas as pd

from mindsdb.integrations.libs.api_handler import APIResource
from mindsdb.integrations.utilities.sql_utils import (
    FilterCondition,
    SortColumn
)


class ListFilesTable(APIResource):
    """
    The table abstraction for querying the files (tables) in Microsoft OneDrive.
    """

    def list(
        self,
        conditions: List[FilterCondition] = None,
        limit: int = None,
        sort: List[SortColumn] = None,
        targets: List[Text] = None,
        **kwargs
    ):
        """
        Lists the files in Microsoft OneDrive.

        Args:
            conditions (List[FilterCondition]): The conditions to filter the files.
            limit (int): The maximum number of files to return.
            sort (List[SortColumn]): The columns to sort the files by.
            targets (List[Text]): The columns to return in the result.

        Returns:
            pd.DataFrame: The list of files in Microsoft OneDrive based on the specified clauses.
        """
        client = self.handler.connect()
        files = client.get_all_items()

        data = []
        for file in files:
            item = {
                "name": file["name"],
                "path": file["path"],
                "extension": file["name"].split(".")[-1]
            }

            # If the 'content' column is explicitly requested, fetch the content of the file.
            if targets and "content" in targets:
                item["content"] = client.get_item_content(file["path"])

            # If a SELECT * query is executed, i.e., targets is empty, set the content to None.
            elif not targets:
                item["content"] = None

            data.append(item)

        df = pd.DataFrame(data)
        return df

    def get_columns(self):
        return ["name", "path", "extension", "content"]


class FileTable(APIResource):
    """
    The table abstraction for querying the content of a file (table) in Microsoft OneDrive.
    """

    def list(self, targets: List[str] = None, table_name=None, *args, **kwargs) -> pd.DataFrame:
        """
        Retrieves the content of the specified file (table) in Microsoft OneDrive.

        Args:
            targets (List[str]): The columns to return in the result.
            table_name (str): The name of the file (table) to retrieve.

        Returns:
            pd.DataFrame: The content of the specified file (table) in Microsoft OneDrive.
        """
        client = self.handler.connect()

        file_content = client.get_item_content(table_name)
        file_extension = table_name.split(".")[-1]

        # Read the file content based and return a DataFrame based on the file extension.
        if file_extension == "csv":
            df = pd.read_csv(BytesIO(file_content))

        elif file_extension == "tsv":
            df = pd.read_csv(BytesIO(file_content), sep="\t")

        elif file_extension == "json":
            df = pd.DataFrame(file_content)

        elif file_extension == "parquet":
            df = pd.read_parquet(BytesIO(file_content))

        return df



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    user={
        'type': ARG_TYPE.STR,
        'description': 'The user name used to authenticate with the Vertica server.'
    },
    password={
        'type': ARG_TYPE.PWD,
        'description': 'The password to authenticate the user with the VERTICA server.',
        'secret': True
    },
    database={
        'type': ARG_TYPE.STR,
        'description': 'The database name to use when connecting with the VERTICA server.'
    },
    host={
        'type': ARG_TYPE.STR,
        'description': 'The host name or IP address of the VERTICA server. NOTE: use \'127.0.0.1\' instead of \'localhost\' to connect to local server.'
    },
    port={
        'type': ARG_TYPE.INT,
        'description': 'The TCP/IP port of the VERTICA server. Must be an integer.'
    },
    schema_name={
        'type': ARG_TYPE.STR,
        'description': 'Table are listed according to schema name (it is optional). Note: Default value is "public"'
    }
)

connection_args_example = OrderedDict(
    host='127.0.0.1',
    port=5433,
    user='root',
    password='password',
    database='database',
    schema_name='xyz'
)



import ast
from datetime import timedelta
import uuid

import pandas as pd
from couchbase.auth import PasswordAuthenticator
from couchbase.cluster import Cluster
from couchbase.exceptions import UnAmbiguousTimeoutException
from couchbase.options import ClusterOptions
from couchbase.exceptions import CouchbaseException
from typing import List, Union

from mindsdb.utilities import log
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE,
)
from mindsdb.integrations.libs.vectordatabase_handler import (
    FilterCondition,
    TableField,
    VectorStoreHandler,
)


logger = log.getLogger(__name__)


class CouchbaseVectorHandler(VectorStoreHandler):
    """
    This handler handles connection and execution of the Couchbase statements.
    """

    name = "couchbasevector"
    DEFAULT_TIMEOUT_SECONDS = 5

    def __init__(self, name, **kwargs):
        super().__init__(name)
        self.connection_data = kwargs.get("connection_data")

        self.scope = self.connection_data.get("scope") or "_default"

        self.bucket_name = self.connection_data.get("bucket")
        self.cluster = None

        self.is_connected = False

    def connect(self):
        """
        Set up connections required by the handler.

        Returns:
            The connected cluster.
        """
        if self.is_connected:
            return self.cluster

        auth = PasswordAuthenticator(
            self.connection_data.get("user"),
            self.connection_data.get("password"),
            # NOTE: If using SSL/TLS, add the certificate path.
            # We strongly reccomend this for production use.
            # cert_path=cert_path
        )

        options = ClusterOptions(auth)

        conn_str = self.connection_data.get("connection_string")
        # wan_development is used to avoid latency issues while connecting to Couchbase over the internet
        options.apply_profile("wan_development")
        # connect to the cluster
        cluster = Cluster(
            conn_str,
            options,
        )

        try:
            # wait until the cluster is ready for use
            cluster.wait_until_ready(
                timedelta(seconds=self.DEFAULT_TIMEOUT_SECONDS)
            )
            self.is_connected = cluster.connected
            self.cluster = cluster
        except UnAmbiguousTimeoutException:
            self.is_connected = False
            raise

        return self.cluster

    def disconnect(self):
        """Close any existing connections
        Should switch self.is_connected.
        """
        if self.is_connected is False:
            return
        self.is_connected = self.cluster.connected
        return

    def check_connection(self) -> StatusResponse:
        """
        Check the connection of the Couchbase bucket
        :return: success status and error message if error occurs
        """
        result = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            cluster = self.connect()
            result.success = cluster.connected
        except UnAmbiguousTimeoutException as e:
            logger.error(
                f'Error connecting to Couchbase {self.connection_data["bucket"]}, {e}!'
            )
            result.error_message = str(e)

        if result.success is True and need_to_close:
            self.disconnect()
        if result.success is False and self.is_connected is True:
            self.is_connected = False
        return result

    def _translate_conditions(
        self, conditions: List[FilterCondition]
    ) -> Union[dict, None]:
        """
        Translate filter conditions to a dictionary
        """
        if conditions is None:
            return {}

        return {
            condition.column: {
                "op": condition.op.value,
                "value": condition.value,
            }
            for condition in conditions
        }

    def _construct_full_after_from_query(
        self,
        where_query: str,
        limit_query: str,
        offset_query: str,
        search_query: str,
    ) -> str:

        return f"{where_query} {search_query} {limit_query} {offset_query} "

    def _construct_where_query(self, filter_conditions=None):
        """
        Construct where querys from filter conditions
        """
        if filter_conditions is None:
            return ""

        where_querys = []
        metadata_conditions = {
            key: value
            for key, value in filter_conditions.items()
            if not key.startswith(TableField.EMBEDDINGS.value)
        }
        for key, value in metadata_conditions.items():
            if value["op"].lower() == "in":
                values = list(repr(i) for i in value["value"])
                value["value"] = "({})".format(", ".join(values))
            else:
                value["value"] = repr(value["value"])
            where_querys.append(f'{key} {value["op"]} {value["value"]}')

        if len(where_querys) > 1:
            return f"WHERE {' AND '.join(where_querys)}"
        elif len(where_querys) == 1:
            return f"WHERE {where_querys[0]}"
        else:
            return ""

    def _construct_search_query(
        self, table_name: str, field: str, vector: list, k: int, condition: str
    ):
        """
        Construct a SEARCH query for KNN
        :param table_name: Name of the table
        :param field: The field on which to perform the search (e.g., embeddings)
        :param vector: The vector to search against
        :param k: The number of nearest neighbors to return, default: 2
        :return: The SEARCH query as a string
        """
        k_value = k if k is not None else 2
        search_query = f"""
        {condition} SEARCH({table_name}, {{
            "fields": ["*"],
            "query": {{
                "match_none": ""
            }},
            "knn": [
            {{
                "k": {k_value},
                "field": "{field}",
                "vector": {vector}
            }}
            ]
        }})
        """
        return search_query.strip()

    def select(
        self,
        table_name: str,
        columns: List[str] = None,
        conditions: List[FilterCondition] = None,
        offset: int = None,
        limit: int = None,
    ) -> pd.DataFrame:
        filter_conditions = self._translate_conditions(conditions)
        cluster = self.connect()
        bucket = cluster.bucket(self.bucket_name)
        scope = bucket.scope(self.scope)
        documents, metadatas, embeddings = [], [], []

        vector_filter = (
            next(
                (
                    condition
                    for condition in conditions
                    if condition.column == TableField.EMBEDDINGS.value
                ),
                None,
            )
            if conditions
            else None
        )
        limit_query = f"LIMIT {limit}" if limit else ""
        offset_query = f"OFFSET {offset}" if offset else ""
        if vector_filter:
            vector = vector_filter.value
            if not isinstance(vector, list):
                vector = ast.literal_eval(vector)

            where_query = self._construct_where_query(filter_conditions)
            if where_query == "":
                search_query = self._construct_search_query(
                    table_name,
                    TableField.EMBEDDINGS.value,
                    vector_filter.value,
                    limit,
                    "WHERE",
                )
            else:
                search_query = self._construct_search_query(
                    table_name,
                    TableField.EMBEDDINGS.value,
                    vector_filter.value,
                    limit,
                    "AND",
                )
            after_from_query = self._construct_full_after_from_query(
                where_query, limit_query, offset_query, search_query
            )

            if columns is None:
                targets = "id, content, embeddings, metadata"
            else:
                targets = ", ".join(columns)
            query = f"SELECT SEARCH_SCORE() AS score, {targets} FROM {table_name} {after_from_query}"
            try:
                result = scope.query(query)
            except CouchbaseException as e:
                raise Exception(f"Error while executing query: '{e}'")

            # Process results
            ids, documents, distances = [], [], []
            for hit in result.rows():
                ids.append(hit.get("id", ""))
                documents.append(hit.get("content", ""))
                embeddings.append(hit.get("embeddings", []))
                metadatas.append(hit.get("metadata", {}))
                distances.append(hit.get("score", ""))
        else:

            where_query = self._construct_where_query(filter_conditions)
            after_from_query = self._construct_full_after_from_query(
                where_query, limit_query, offset_query, ""
            )

            if columns is None:
                targets = "id, content, embeddings, metadata"
            else:
                targets = ", ".join(columns)

            query = f"SELECT {targets} FROM {table_name} {after_from_query}"
            try:
                result = scope.query(query)
            except CouchbaseException as e:
                raise Exception(f"Error while executing query: '{e}'")

            ids = []
            documents = []
            for hit in result.rows():
                ids.append(hit.get("id", ""))
                documents.append(hit.get("content", ""))
                embeddings.append(hit.get("embeddings", []))
                metadatas.append(hit.get("metadata", {}))

            distances = None

        # Prepare the payload
        payload = {
            TableField.ID.value: ids,
            TableField.CONTENT.value: [doc for doc in documents],
            TableField.METADATA.value: [doc for doc in metadatas],
            TableField.EMBEDDINGS.value: [doc for doc in embeddings],
        }
        if columns:
            payload = {
                column: payload[column]
                for column in columns
                if column in payload
            }
        if distances is not None:
            payload[TableField.DISTANCE.value] = distances
        return pd.DataFrame(payload)

    def insert(self, table_name: str, data: pd.DataFrame) -> Response:
        """
        Insert data into Couchbase.
        """

        data.dropna(axis=1, inplace=True)
        # Convert DataFrame to list of dictionaries
        records = data.to_dict(orient="records")
        cluster = self.connect()
        bucket = cluster.bucket(self.bucket_name)
        scope = bucket.scope(self.scope)
        collection = scope.collection(table_name)

        for record in records:
            doc_id = record.get(TableField.ID.value, str(uuid.uuid4()))
            document = {TableField.ID.value: doc_id}

            if TableField.CONTENT.value in record:
                document[TableField.CONTENT.value] = record[
                    TableField.CONTENT.value
                ]

            if TableField.EMBEDDINGS.value in record:
                document[TableField.EMBEDDINGS.value] = record[
                    TableField.EMBEDDINGS.value
                ]
                if not isinstance(document[TableField.EMBEDDINGS.value], list):
                    document[TableField.EMBEDDINGS.value] = ast.literal_eval(
                        document[TableField.EMBEDDINGS.value]
                    )

            if TableField.METADATA.value in record:
                document[TableField.METADATA.value] = record[
                    TableField.METADATA.value
                ]
            document_key = f"{table_name}::{doc_id}"

            collection.upsert(document_key, document)
        return Response(resp_type=RESPONSE_TYPE.OK)

    def upsert(self, table_name: str, data: pd.DataFrame):
        return self.insert(table_name, data)

    def update(
        self,
        table_name: str,
        data: pd.DataFrame,
        key_columns: List[str] = None,
    ):
        """
        Update data in Couchbase.
        """
        # Convert DataFrame to list of dictionaries
        records = data.to_dict(orient="records")
        cluster = self.connect()
        bucket = cluster.bucket(self.bucket_name)
        scope = bucket.scope(self.scope)
        collection = scope.collection(table_name)
        try:
            for record in records:
                doc_id = record.get(TableField.ID.value)
                if doc_id:
                    existing_doc = self.collection.get(doc_id)
                    if existing_doc:
                        updated_doc = existing_doc.content
                        updated_doc.update(record)
                        collection.replace(doc_id, updated_doc)
        except CouchbaseException as e:
            raise Exception(f"Error while updating document: '{e}'")

    def delete(
        self, table_name: str, conditions: List[FilterCondition] = None
    ):
        """
        Delete documents in Couchbase based on conditions.
        """
        filter_conditions = self._translate_conditions(conditions)
        cluster = self.connect()
        bucket = cluster.bucket(self.bucket_name)
        scope = bucket.scope(self.scope)
        where_query = self._construct_where_query(filter_conditions)

        query = f"DELETE FROM {table_name} {where_query}"
        try:
            _ = scope.query(query)
        except CouchbaseException as e:
            raise Exception(
                f"Error while performing delete query index: '{e}'"
            )

    def create_table(self, table_name: str, if_not_exists=True):
        """
        In Couchbase, tables are represented as collections within a bucket.
        This method creates a new collection if it doesn't exist.
        """
        cluster = self.connect()
        bucket = cluster.bucket(self.bucket_name)
        try:
            bucket.collections().create_collection(
                scope_name=self.scope, collection_name=table_name
            )
        except Exception as e:
            raise Exception(f"Error while creating table: '{e}'")

    def drop_table(self, table_name: str, if_exists=True) -> Response:
        """
        Drop a collection in Couchbase.
        """
        cluster = self.connect()
        bucket = cluster.bucket(self.bucket_name)
        scope = bucket.scope(self.scope)
        _ = scope.collection(table_name)
        try:
            bucket.collections().drop_collection(table_name)
            return Response(resp_type=RESPONSE_TYPE.OK)
        except Exception as e:
            return Response(resp_type=RESPONSE_TYPE.ERROR, error_message=e)

    def get_tables(self) -> Response:
        """
        Get the list of collections in the Couchbase bucket.
        """
        cluster = self.connect()
        bucket = cluster.bucket(self.bucket_name)
        collections = bucket.collections().get_all_scopes()
        collection_names = [
            coll.name for scope in collections for coll in scope.collections
        ]
        collections_df = pd.DataFrame(
            columns=["table_name"], data=collection_names
        )
        return Response(
            resp_type=RESPONSE_TYPE.TABLE, data_frame=collections_df
        )

    def get_columns(self, table_name: str) -> Response:
        """
        Get the columns (fields) of a Couchbase collection.
        """
        try:
            cluster = self.connect()
            bucket = cluster.bucket(self.bucket_name)
            scope = bucket.scope(self.scope)
            _ = scope.collection(table_name)
        except Exception:
            return Response(
                resp_type=RESPONSE_TYPE.ERROR,
                error_message=f"Table {table_name} does not exist!",
            )
        return super().get_columns(table_name)



from mindsdb.integrations.handlers.huggingface_handler.finetune import (
    _finetune_cls,
    _finetune_fill_mask,
    _finetune_question_answering,
    _finetune_summarization,
    _finetune_text_generation,
    _finetune_translate,
)

# todo once we have moved predict tasks functions into a separate function
# PREDICT_MAP = {
#             'text-classification': self.predict_text_classification,
#             'zero-shot-classification': self.predict_zero_shot,
#             'translation': self.predict_translation,
#             'summarization': self.predict_summarization,
#             'fill-mask': self.predict_fill_mask
#         }

FINETUNE_MAP = {
    "text-classification": _finetune_cls,
    "zero-shot-classification": _finetune_cls,
    "translation": _finetune_translate,
    "summarization": _finetune_summarization,
    "fill-mask": _finetune_fill_mask,
    "text-generation": _finetune_text_generation,
    "question-answering": _finetune_question_answering,
}
